from __future__ import annotations

import dataclasses
import textwrap
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from dissect.target.target import Event, Target

if TYPE_CHECKING:
    import argparse

    from dissect.target.plugin import FunctionDescriptor, Plugin, PluginRegistry

BLOCK_INDENT = 4 * " "


@dataclass
class TargetExecutionReport:
    target: Target

    incompatible_plugins: set[str] = dataclasses.field(default_factory=set)
    registered_plugins: set[str] = dataclasses.field(default_factory=set)

    func_errors: dict[str, str] = dataclasses.field(default_factory=dict)
    func_execs: set[str] = dataclasses.field(default_factory=set)

    def add_incompatible_plugin(self, plugin_name: str) -> None:
        self.incompatible_plugins.add(plugin_name)

    def add_registered_plugin(self, plugin_name: str) -> None:
        self.registered_plugins.add(plugin_name)

    def add_func_error(self, func: str, stacktrace: str) -> None:
        self.func_errors[func] = stacktrace

    def as_dict(self) -> dict[str, Any]:
        return {
            "target": str(self.target),
            "incompatible_plugins": sorted(self.incompatible_plugins),
            "registered_plugins": sorted(self.registered_plugins),
            "func_errors": self.func_errors,
            "func_execs": sorted(self.func_execs),
        }


@dataclass
class ExecutionReport:
    plugin_import_errors: dict[str, str] = dataclasses.field(default_factory=dict)

    target_reports: list[TargetExecutionReport] = dataclasses.field(default_factory=list)

    cli_args: dict[str, Any] = dataclasses.field(default_factory=dict)

    def set_cli_args(self, args: argparse.Namespace) -> None:
        args = ((key, str(value)) for (key, value) in vars(args).items())
        self.cli_args.update(args)

    def set_plugin_stats(self, plugins: PluginRegistry) -> None:
        for details in plugins.__failed__:
            self.plugin_import_errors[details.module] = "".join(details.stacktrace)

    def get_formatted_report(self) -> str:
        blocks = [
            make_cli_args_overview(self),
            make_plugin_import_errors_overview(self),
            *[format_target_report(report) for report in self.target_reports],
        ]
        return "\n".join(blocks)

    def add_target_report(self, target: Target) -> TargetExecutionReport:
        target_report = TargetExecutionReport(target=target)
        self.target_reports.append(target_report)
        return target_report

    def get_target_report(self, target: Target, create: bool = False) -> TargetExecutionReport:
        target_report = next(filter(lambda r: r.target == target, self.target_reports), None)
        if target_report is None and create:
            target_report = self.add_target_report(target)
        return target_report

    @staticmethod
    def _get_plugin_name(plugin_cls: type[Plugin]) -> str:
        return f"{plugin_cls.__module__}.{plugin_cls.__qualname__}"

    def log_incompatible_plugin(
        self,
        target: Target,
        _: Event,
        plugin_cls: type[Plugin] | None = None,
        plugin_desc: FunctionDescriptor | None = None,
    ) -> None:
        if not plugin_cls and not plugin_desc:
            raise ValueError("Either `plugin_cls` or `plugin_desc` must be set")

        target_report = self.get_target_report(target, create=True)

        if plugin_cls:
            plugin_name = self._get_plugin_name(plugin_cls)
        elif plugin_desc:
            plugin_name = f"{plugin_desc.module}.{plugin_desc.qualname}"

        target_report.add_incompatible_plugin(plugin_name)

    def log_registered_plugin(self, target: Target, _: Event, plugin_inst: Plugin) -> None:
        target_report = self.get_target_report(target, create=True)
        plugin_cls = type(plugin_inst)
        plugin_name = self._get_plugin_name(plugin_cls)
        target_report.add_registered_plugin(plugin_name)

    def log_func_error(self, target: Target, _: Event, func: str, stacktrace: str) -> None:
        target_report = self.get_target_report(target, create=True)
        target_report.add_func_error(func, stacktrace)

    def log_func_execution(self, target: Target, _: Event, func: str) -> None:
        target_report = self.get_target_report(target, create=True)
        target_report.func_execs.add(func)

    def set_event_callbacks(self, target_cls: type[Target]) -> None:
        target_cls.set_event_callback(
            event_type=Event.INCOMPATIBLE_PLUGIN,
            event_callback=self.log_incompatible_plugin,
        )
        target_cls.set_event_callback(
            event_type=Event.REGISTERED_PLUGIN,
            event_callback=self.log_registered_plugin,
        )
        target_cls.set_event_callback(
            event_type=Event.FUNC_EXEC,
            event_callback=self.log_func_execution,
        )
        target_cls.set_event_callback(
            event_type=Event.FUNC_EXEC_ERROR,
            event_callback=self.log_func_error,
        )

    def as_dict(self) -> dict[str, Any]:
        return {
            "plugin_import_errors": self.plugin_import_errors,
            "target_reports": [report.as_dict() for report in self.target_reports],
            "cli_args": self.cli_args,
        }


def make_cli_args_overview(report: ExecutionReport) -> str:
    header = "CLI arguments:"

    rows = []
    for key, value in sorted(report.cli_args.items()):
        rows.append(f"{key}: {value}")

    block = "\n".join(rows)
    block = textwrap.indent(block, prefix=BLOCK_INDENT)
    return f"{header}\n{block}"


def make_plugin_import_errors_overview(report: ExecutionReport, short: bool = True) -> str:
    header = "Plugin import errors:"

    rows = []
    for module, trace in sorted(report.plugin_import_errors.items()):
        trace = trace.rstrip()
        trace = textwrap.indent(trace, prefix=BLOCK_INDENT)
        rows.append(f"{module}:\n{trace}")

    block = "\n".join(rows)
    block = textwrap.indent(block, prefix=BLOCK_INDENT)
    return f"{header}\n{block}"


def format_target_report(target_report: TargetExecutionReport) -> str:
    blocks = [f"Target: {target_report.target}"]

    registered_plugins_header = "Registered plugins:"
    registered_plugins_rows = sorted(target_report.registered_plugins)
    registered_plugins_rows_block = textwrap.indent(
        "\n".join(registered_plugins_rows),
        prefix=BLOCK_INDENT,
    )
    registered_plugins_block = textwrap.indent(
        f"{registered_plugins_header}\n{registered_plugins_rows_block}",
        prefix=BLOCK_INDENT,
    )
    blocks.append(registered_plugins_block)

    incompatible_plugins_header = "Incompatible plugins:"
    incompatible_plugins_rows = sorted(target_report.incompatible_plugins)
    incompatible_plugins_rows_block = textwrap.indent(
        "\n".join(incompatible_plugins_rows),
        prefix=BLOCK_INDENT,
    )
    incompatible_plugins_block = textwrap.indent(
        f"{incompatible_plugins_header}\n{incompatible_plugins_rows_block}",
        prefix=BLOCK_INDENT,
    )
    blocks.append(incompatible_plugins_block)

    func_errors_header = "Function errors:"
    func_errors_rows = []
    for func, stacktrace in sorted(target_report.func_errors.items()):
        stacktrace = textwrap.indent(stacktrace, prefix=BLOCK_INDENT)
        func_errors_rows.append(f"{func}:\n{stacktrace}")

    func_errors_rows_block = textwrap.indent("\n".join(func_errors_rows), prefix=BLOCK_INDENT)
    func_errors_block = textwrap.indent(
        f"{func_errors_header}\n{func_errors_rows_block}",
        prefix=BLOCK_INDENT,
    )
    blocks.append(func_errors_block)

    func_execs_header = "Function executions:"
    func_execs_rows = sorted(target_report.func_execs)
    func_execs_rows_block = textwrap.indent("\n".join(func_execs_rows), prefix=BLOCK_INDENT)
    func_execs_block = textwrap.indent(
        f"{func_execs_header}\n{func_execs_rows_block}",
        prefix=BLOCK_INDENT,
    )
    blocks.append(func_execs_block)

    return "\n".join(blocks)



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from dissect.esedb.c_esedb import decode_guid
from dissect.esedb.esedb import EseDB
from dissect.util.ts import oatimestamp

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

    from dissect.target.target import Target

WuaHistoryRecord = TargetRecordDescriptor(
    "filesystem/windows/wua_history",
    [
        ("datetime", "ts"),
        ("varint", "id_event"),
        ("varint", "status_id"),
        ("varint", "server_selection"),
        ("string", "mapped_result"),
        ("string", "unmapped_result"),
        ("string", "update_id"),
        ("string", "server_id"),
        ("varint", "flags"),
        ("string", "client_id"),
        ("string", "title"),
        ("string", "description"),
        ("string", "uninstall_notes"),
        ("string", "support_url"),
        ("string", "uninstall_steps"),
        ("string", "categories"),
        ("string", "more_info_url"),
        ("string", "path"),
        ("varint", "id_user"),
        ("string", "is_service_is_additional"),
        ("string", "classification"),
        # Enriched fields
        ("string", "kb"),
        ("string", "status_mapped"),
        ("string", "mapped_result_string"),
        ("string", "mapped_result_description"),
        ("string", "unmapped_result_string"),
        ("string", "unmapped_result_description"),
        ("string", "server_id_mapped"),
        ("string", "server_selection_mapped"),
        ("string", "classification_mapped"),
    ],
)

TBHISTORY_COLUMN_MAP = {
    "IdEvent": "id_event",
    "Status": "status_id",
    "ServerSelection": "server_selection",
    "MappedResult": "mapped_result",
    "UnmappedResult": "unmapped_result",
    "UpdateId": "update_id",
    "ServerId": "server_id",
    "Date": "ts",
    "Flags": "flags",
    "ClientId": "client_id",
    "Title": "title",
    "Description": "description",
    "UninstallNotes": "uninstall_notes",
    "SupportUrl": "support_url",
    "UninstallSteps": "uninstall_steps",
    "Categories": "categories",
    "MoreInfoUrl": "more_info_url",
    "IdUser": "id_user",
    "IsServiceIsAdditional": "is_service_is_additional",
    "Classification": "classification",
    # Enriched fields
    "kb": "kb",
    "status_mapped": "status_mapped",
    "mapped_result_string": "mapped_result_string",
    "mapped_result_description": "mapped_result_description",
    "unmapped_result_string": "unmapped_result_string",
    "unmapped_result_description": "unmapped_result_description",
    "server_id_mapped": "server_id_mapped",
    "server_selection_mapped": "server_selection_mapped",
    "classification_mapped": "classification_mapped",
}

CLASSIFICATION_MAP = {
    "5c9376ab-8ce6-464a-b136-22113dd69801": "Application",
    "434de588-ed14-48f5-8eed-a15e09a991f6": "Connectors",
    "e6cf1350-c01b-414d-a61f-263d14d133b4": "CriticalUpdates",
    "e0789628-ce08-4437-be74-2495b842f43b": "DefinitionUpdates",
    "e140075d-8433-45c3-ad87-e72345b36078": "DeveloperKits",
    "b54e7d24-7add-428f-8b75-90a396fa584f": "FeaturePacks",
    "9511d615-35b2-47bb-927f-f73d8e9260bb": "Guidance",
    "0fa1201d-4330-4fa8-8ae9-b877473b6441": "SecurityUpdates",
    "68c5b0a3-d1a6-4553-ae49-01d3a7827828": "ServicePacks",
    "b4832bd8-e735-4761-8daf-37f882276dab": "Tools",
    "28bc880e-0592-4cbf-8f95-c79b17911d5f": "UpdateRollups",
    "cd5ffd1e-e932-4e3a-bf74-18bf0b1bbd83": "Updates",
    "3689bdc8-b205-4af4-8d4a-a63924c5e9d5": "Upgrades",
    "ebfc1fc5-71a4-4f7b-9aca-3b9a503104a0": "Drivers",
}

SERVER_ID_MAP = {
    "00000000-0000-0000-0000-000000000000": "Unspecified / Default",
    "9482f4b4-e343-43b6-b170-9a65bc822c77": "Windows Update",
    "7971f918-a847-4430-9279-4a52d1efe18d": "Microsoft Update",
    "855e8a7c-ecb4-4ca3-b045-1dfa50104289": "Store",
    "8b24b027-1dee-babb-9a95-3517dfb9c552": "OS Flighting",
    "3da21691-e39d-4da6-8a4b-b43877bcb1b7": "WSUS or Configuration Manager",
}

SERVER_SELECTION_MAP = {
    0: "ssDefault",
    1: "ssManagedServer",
    2: "ssWindowsUpdate",
    3: "ssOthers",
}

STATUS_MAP = {
    1: "Success",
    2: "Failure",
}

WUA_CODE_MAP = {
    "0x0": ["Success", "Success"],
    "0x240001": ["WU_S_SERVICE_STOP", "Windows Update Agent was stopped successfully."],
    "0x240002": ["WU_S_SELFUPDATE", "Windows Update Agent updated itself."],
    "0x240003": ["WU_S_UPDATE_ERROR", "Operation completed successfully but there were errors applying the updates."],
    "0x240004": [
        "WU_S_MARKED_FOR_DISCONNECT",
        "A callback was marked to be disconnected later because the request to disconnect the operation came while a"
        " callback was executing.",
    ],
    "0x240005": ["WU_S_REBOOT_REQUIRED", "The system must be restarted to complete installation of the update."],
    "0x240006": ["WU_S_ALREADY_INSTALLED", "The update to be installed is already installed on the system."],
    "0x240007": ["WU_S_ALREADY_UNINSTALLED", "The update to be removed is not installed on the system."],
    "0x240008": ["WU_S_ALREADY_DOWNLOADED", "The update to be downloaded has already been downloaded."],
    "0x8024400A": [
        "WU_E_PT_SOAPCLIENT_PARSE",
        "SOAPCLIENT_PARSE_ERROR - SOAP client failed to parse the response from the server.",
    ],
    "0x8024600a": [
        "WU_E_DM_DOWNLOADLOCATIONCHANGED",
        "A download must be restarted because the location of the source of the download has changed.",
    ],
    "0x8024D008": [
        "WU_E_SELFUPDATE_SKIP_ON_FAILURE",
        "An update to the Windows Update Agent was skipped because previous attempts to update have failed.",
    ],
    "0x8024D011": ["WU_E_SELFUPDATE_REQUIRED", "Windows Update Agent must be updated before search can continue."],
    "0x8024D012": [
        "WU_E_SELFUPDATE_REQUIRED_ADMIN",
        "Windows Update Agent must be updated before search can continue. An administrator is required to perform the "
        "operation",
    ],
    "0x80243FFF": [
        "WU_E_AUCLIENT_UNEXPECTED",
        "There was a user interface error not covered by another WU_E_AUCLIENT_* error code.",
    ],
    "0x8024A000": ["WU_E_AU_NOSERVICE", "Automatic Updates was unable to service incoming requests."],
    "0x8024A002": [
        "WU_E_AU_NONLEGACYSERVER",
        "The old version of the Automatic Updates client has stopped because the WSUS server has been upgraded.",
    ],
    "0x8024A003": ["WU_E_AU_LEGACYCLIENTDISABLED", "The old version of the Automatic Updates client was disabled."],
    "0x8024A004": [
        "WU_E_AU_PAUSED",
        "Automatic Updates was unable to process incoming requests because it was paused.",
    ],
    "0x8024A005": ["WU_E_AU_NO_REGISTERED_SERVICE", "No unmanaged service is registered with AU."],
    "0x8024AFFF": ["WU_E_AU_UNEXPECTED", "An Automatic Updates error not covered by another WU_E_AU * code."],
    "0x80243001": [
        "WU_E_INSTALLATION_RESULTS_UNKNOWN_VERSION",
        "The results of download and installation could not be read from the registry due to an unrecognized data "
        "format version.",
    ],
    "0x80243002": [
        "WU_E_INSTALLATION_RESULTS_INVALID_DATA",
        "The results of download and installation could not be read from the registry due to an invalid data format.",
    ],
    "0x80243003": [
        "WU_E_INSTALLATION_RESULTS_NOT_FOUND",
        "The results of download and installation are not available; the operation may have failed to start.",
    ],
    "0x80243004": [
        "WU_E_TRAYICON_FAILURE",
        "A failure occurred when trying to create an icon in the taskbar notification area.",
    ],
    "0x80243FFD": [
        "WU_E_NON_UI_MODE",
        "Unable to show UI when in non-UI mode; WU client UI modules may not be installed.",
    ],
    "0x80243FFE": ["WU_E_WUCLTUI_UNSUPPORTED_VERSION", "Unsupported version of WU client UI exported functions."],
    "0x8024043D": ["WU_E_SERVICEPROP_NOTAVAIL", "The requested service property isn't available."],
    "0x80249001": ["WU_E_INVENTORY_PARSEFAILED", "Parsing of the rule file failed."],
    "0x80249002": [
        "WU_E_INVENTORY_GET_INVENTORY_TYPE_FAILED",
        "Failed to get the requested inventory type from the server.",
    ],
    "0x80249003": ["WU_E_INVENTORY_RESULT_UPLOAD_FAILED", "Failed to upload inventory result to the server."],
    "0x80249004": ["WU_E_INVENTORY_UNEXPECTED", "There was an inventory error not covered by another error code."],
    "0x80249005": [
        "WU_E_INVENTORY_WMI_ERROR",
        "A WMI error occurred when enumerating the instances for a particular class.",
    ],
    "0x8024E001": [
        "WU_E_EE_UNKNOWN_EXPRESSION",
        "An expression evaluator operation could not be completed because an expression was unrecognized.",
    ],
    "0x8024E002": [
        "WU_E_EE_INVALID_EXPRESSION",
        "An expression evaluator operation could not be completed because an expression was invalid.",
    ],
    "0x8024E003": [
        "WU_E_EE_MISSING_METADATA",
        "An expression evaluator operation could not be completed because an expression contains an incorrect number "
        "of metadata nodes.",
    ],
    "0x8024E004": [
        "WU_E_EE_INVALID_VERSION",
        "An expression evaluator operation could not be completed because the version of the serialized expression "
        "data is invalid.",
    ],
    "0x8024E005": ["WU_E_EE_NOT_INITIALIZED", "The expression evaluator could not be initialized."],
    "0x8024E006": [
        "WU_E_EE_INVALID_ATTRIBUTEDATA",
        "An expression evaluator operation could not be completed because there was an invalid attribute.",
    ],
    "0x8024E007": [
        "WU_E_EE_CLUSTER_ERROR",
        "An expression evaluator operation could not be completed because the cluster state of the computer could not "
        "be determined.",
    ],
    "0x8024EFFF": [
        "WU_E_EE_UNEXPECTED",
        "There was an expression evaluator error not covered by another WU_E_EE_* error code.",
    ],
    "0x80247001": [
        "WU_E_OL_INVALID_SCANFILE",
        "An operation could not be completed because the scan package was invalid.",
    ],
    "0x80247002": [
        "WU_E_OL_NEWCLIENT_REQUIRED",
        "An operation could not be completed because the scan package requires a greater version of the Windows Update "
        "Agent.",
    ],
    "0x80247FFF": ["WU_E_OL_UNEXPECTED", "Search using the scan package failed."],
    "0x8024F001": ["WU_E_REPORTER_EVENTCACHECORRUPT", "The event cache file was defective."],
    "0x8024F002": [
        "WU_E_REPORTER_",
        "EVENTNAMESPACEPARSEFAILEDThe XML in the event namespace descriptor could not be parsed.",
    ],
    "0x8024F003": ["WU_E_INVALID_EVENT", "The XML in the event namespace descriptor could not be parsed."],
    "0x8024F004": ["WU_E_SERVER_BUSY", "The server rejected an event because the server was too busy."],
    "0x8024FFFF": ["WU_E_REPORTER_UNEXPECTED", "There was a reporter error not covered by another error code."],
    "0x80245001": ["WU_E_REDIRECTOR_LOAD_XML", "The redirector XML document could not be loaded into the DOM class."],
    "0x80245002": ["WU_E_REDIRECTOR_S_FALSE", "The redirector XML document is missing some required information."],
    "0x80245003": [
        "WU_E_REDIRECTOR_ID_SMALLER",
        "The redirector ID in the downloaded redirector cab is less than in the cached cab.",
    ],
    "0x80245FFF": [
        "WU_E_REDIRECTOR_UNEXPECTED",
        "The redirector failed for reasons not covered by another WU_E_REDIRECTOR_* error code.",
    ],
    "0x80244000": [
        "WU_E_PT_SOAPCLIENT_BASE",
        "WU_E_PT_SOAPCLIENT_* error codes map to the SOAPCLIENT_ERROR enum of the ATL Server Library.",
    ],
    "0x80244001": [
        "WU_E_PT_SOAPCLIENT_INITIALIZE",
        "SOAPCLIENT_INITIALIZE_ERROR - initialization of the SOAP client failed, possibly because of an MSXML "
        "installation failure.",
    ],
    "0x80244002": [
        "WU_E_PT_SOAPCLIENT_OUTOFMEMORY",
        "SOAPCLIENT_OUTOFMEMORY - SOAP client failed because it ran out of memory.",
    ],
    "0x80244003": [
        "WU_E_PT_SOAPCLIENT_GENERATE",
        "SOAPCLIENT_GENERATE_ERROR - SOAP client failed to generate the request.",
    ],
    "0x80244004": [
        "WU_E_PT_SOAPCLIENT_CONNECT",
        "SOAPCLIENT_CONNECT_ERROR - SOAP client failed to connect to the server.",
    ],
    "0x80244005": [
        "WU_E_PT_SOAPCLIENT_SEND",
        "SOAPCLIENT_SEND_ERROR - SOAP client failed to send a message for reasons of WU_E_WINHTTP_* error codes.",
    ],
    "0x80244006": [
        "WU_E_PT_SOAPCLIENT_SERVER",
        "SOAPCLIENT_SERVER_ERROR - SOAP client failed because there was a server error.",
    ],
    "0x80244007": [
        "WU_E_PT_SOAPCLIENT_SOAPFAULT",
        "SOAPCLIENT_SOAPFAULT - SOAP client failed because there was a SOAP fault for reasons of WU_E_PT_SOAP_* error "
        "codes.",
    ],
    "0x80244008": [
        "WU_E_PT_SOAPCLIENT_PARSEFAULT",
        "SOAPCLIENT_PARSEFAULT_ERROR - SOAP client failed to parse a SOAP fault.",
    ],
    "0x80244009": [
        "WU_E_PT_SOAPCLIENT_READ",
        "SOAPCLIENT_READ_ERROR - SOAP client failed while reading the response from the server.",
    ],
    "x8024400A": [
        "WU_E_PT_SOAPCLIENT_PARSE",
        "Same as SOAPCLIENT_PARSE_ERROR - SOAP client failed to parse the response from the server.",
    ],
    "0x8024400B": [
        "WU_E_PT_SOAP_VERSION",
        "SOAP_E_VERSION_MISMATCH - SOAP client found an unrecognizable namespace for the SOAP envelope.",
    ],
    "0x8024400C": [
        "WU_E_PT_SOAP_MUST_UNDERSTAND",
        "SOAP_E_MUST_UNDERSTAND - SOAP client was unable to understand a header.",
    ],
    "0x8024400D": [
        "WU_E_PT_SOAP_CLIENT",
        "SOAP_E_CLIENT - SOAP client found the message was malformed; fix before resending.",
    ],
    "0x8024400E": [
        "WU_E_PT_SOAP_SERVER",
        "SOAP_E_SERVER - The SOAP message could not be processed due to a server error; resend later.",
    ],
    "0x8024400F": ["WU_E_PT_WMI_ERROR", "There was an unspecified Windows Management Instrumentation (WMI) error."],
    "0x80244010": [
        "WU_E_PT_EXCEEDED_MAX_SERVER_TRIPS",
        "The number of round trips to the server exceeded the maximum limit.",
    ],
    "0x80244011": ["WU_E_PT_SUS_SERVER_NOT_SET", "WUServer policy value is missing in the registry."],
    "0x80244012": [
        "WU_E_PT_DOUBLE_INITIALIZATION",
        "Initialization failed because the object was already initialized.",
    ],
    "0x80244013": ["WU_E_PT_INVALID_COMPUTER_NAME", "The computer name could not be determined."],
    "0x80244015": [
        "WU_E_PT_REFRESH_CACHE_REQUIRED",
        "The reply from the server indicates that the server was changed or the cookie was invalid; refresh the state "
        "of the internal cache and retry.",
    ],
    "0x80244016": [
        "WU_E_PT_HTTP_STATUS_BAD_REQUEST",
        "HTTP 400 - the server could not process the request due to invalid syntax.",
    ],
    "0x80244017": ["WU_E_PT_HTTP_STATUS_DENIED", "HTTP 401 - the requested resource requires user authentication."],
    "0x80244018": [
        "WU_E_PT_HTTP_STATUS_FORBIDDEN",
        "HTTP 403 - server understood the request, but declined to fulfill it.",
    ],
    "0x80244019": [
        "WU_E_PT_HTTP_STATUS_NOT_FOUND",
        "HTTP 404 - the server cannot find the requested URI (Uniform Resource Identifier).",
    ],
    "0x8024401A": ["WU_E_PT_HTTP_STATUS_BAD_METHOD", "HTTP 405 - the HTTP method is not allowed."],
    "0x8024401B": ["WU_E_PT_HTTP_STATUS_PROXY_AUTH_REQ", "HTTP 407 - proxy authentication is required."],
    "0x8024401C": ["WU_E_PT_HTTP_STATUS_REQUEST_TIMEOUT", "HTTP 408 - the server timed out waiting for the request."],
    "0x8024401D": [
        "WU_E_PT_HTTP_STATUS_CONFLICT",
        "HTTP 409 - the request was not completed due to a conflict with the current state of the resource.",
    ],
    "0x8024401E": ["WU_E_PT_HTTP_STATUS_GONE", "HTTP 410 - requested resource is no longer available at the server."],
    "0x8024401F": [
        "WU_E_PT_HTTP_STATUS_SERVER_ERROR",
        "HTTP 500 - an error internal to the server prevented fulfilling the request.",
    ],
    "0x80244020": [
        "WU_E_PT_HTTP_STATUS_NOT_SUPPORTED",
        "HTTP 501 - server does not support the functionality required to fulfill the request.",
    ],
    "0x80244021": [
        "WU_E_PT_HTTP_STATUS_BAD_GATEWAY",
        "HTTP 502 - the server, while acting as a gateway or proxy, received an invalid response from the upstream "
        "server it accessed in attempting to fulfill the request.",
    ],
    "0x80244022": ["WU_E_PT_HTTP_STATUS_SERVICE_UNAVAIL", "HTTP 503 - the service is temporarily overloaded."],
    "0x80244023": [
        "WU_E_PT_HTTP_STATUS_GATEWAY_TIMEOUT",
        "HTTP 504 - the request was timed out waiting for a gateway.",
    ],
    "0x80244024": [
        "WU_E_PT_HTTP_STATUS_VERSION_NOT_SUP",
        "HTTP 505 - the server does not support the HTTP protocol version used for the request.",
    ],
    "0x80244025": [
        "WU_E_PT_FILE_LOCATIONS_CHANGED",
        "Operation failed due to a changed file location; refresh internal state and resend.",
    ],
    "0x80244026": [
        "WU_E_PT_REGISTRATION_NOT_SUPPORTED",
        "Operation failed because Windows Update Agent does not support registration with a non-WSUS server.",
    ],
    "0x80244027": [
        "WU_E_PT_NO_AUTH_PLUGINS_REQUESTED",
        "The server returned an empty authentication information list.",
    ],
    "0x80244028": [
        "WU_E_PT_NO_AUTH_COOKIES_CREATED",
        "Windows Update Agent was unable to create any valid authentication cookies.",
    ],
    "0x80244029": ["WU_E_PT_INVALID_CONFIG_PROP", "A configuration property value was wrong."],
    "0x8024402A": ["WU_E_PT_CONFIG_PROP_MISSING", "A configuration property value was missing."],
    "0x8024402B": [
        "WU_E_PT_HTTP_STATUS_NOT_MAPPED",
        "The HTTP request could not be completed and the reason did not correspond to any of the WU_E_PT_HTTP_* error "
        "codes.",
    ],
    "0x8024402C": [
        "WU_E_PT_WINHTTP_NAME_NOT_RESOLVED",
        "ERROR_WINHTTP_NAME_NOT_RESOLVED - the proxy server or target server name cannot be resolved.",
    ],
    "0x8024402F": ["WU_E_PT_ECP_SUCCEEDED_WITH_ERRORS", "External cab file processing completed with some errors."],
    "0x80244030": ["WU_E_PT_ECP_INIT_FAILED", "The external cab processor initialization did not complete."],
    "0x80244031": ["WU_E_PT_ECP_INVALID_FILE_FORMAT", "The format of a metadata file was invalid."],
    "0x80244032": ["WU_E_PT_ECP_INVALID_METADATA", "External cab processor found invalid metadata."],
    "0x80244033": [
        "WU_E_PT_ECP_FAILURE_TO_EXTRACT_DIGEST",
        "The file digest could not be extracted from an external cab file.",
    ],
    "0x80244034": ["WU_E_PT_ECP_FAILURE_TO_DECOMPRESS_CAB_FILE", "An external cab file could not be decompressed."],
    "0x80244035": ["WU_E_PT_ECP_FILE_LOCATION_ERROR", "External cab processor was unable to get file locations."],
    "0x80244FFF": ["WU_E_PT_UNEXPECTED", "A communication error not covered by another WU_E_PT_* error code"],
    "0x8024502D": [
        "WU_E_PT_SAME_REDIR_ID",
        "Windows Update Agent failed to download a redirector cabinet file with a new redirector ID value from the "
        "server during the recovery.",
    ],
    "0x8024502E": [
        "WU_E_PT_NO_MANAGED_RECOVER",
        "A redirector recovery action did not complete because the server is managed.",
    ],
    "0x80246001": [
        "WU_E_DM_URLNOTAVAILABLE",
        "A download manager operation could not be completed because the requested file does not have a URL.",
    ],
    "0x80246002": [
        "WU_E_DM_INCORRECTFILEHASH",
        "A download manager operation could not be completed because the file digest was not recognized.",
    ],
    "0x80246003": [
        "WU_E_DM_UNKNOWNALGORITHM",
        "A download manager operation could not be completed because the file metadata requested an unrecognized hash "
        "algorithm.",
    ],
    "0x80246004": [
        "WU_E_DM_NEEDDOWNLOADREQUEST",
        "An operation could not be completed because a download request is required from the download handler.",
    ],
    "0x80246005": [
        "WU_E_DM_NONETWORK",
        "A download manager operation could not be completed because the network connection was unavailable.",
    ],
    "0x80246006": [
        "WU_E_DM_WRONGBITSVERSION",
        "A download manager operation could not be completed because the version of Background Intelligent Transfer "
        "Service (BITS) is incompatible.",
    ],
    "0x80246007": ["WU_E_DM_NOTDOWNLOADED", "The update has not been downloaded."],
    "0x80246008": [
        "WU_E_DM_FAILTOCONNECTTOBITS",
        "A download manager operation failed because the download manager was unable to connect the Background "
        "Intelligent Transfer Service (BITS).",
    ],
    "0x80246009": [
        "WU_E_DM_BITSTRANSFERERROR",
        "A download manager operation failed because there was an unspecified Background Intelligent Transfer Service "
        "(BITS) transfer error.",
    ],
    "0x8024600A": [
        "WU_E_DM_DOWNLOADLOCATIONCHANGED",
        "A download must be restarted because the location of the source of the download has changed.",
    ],
    "0x8024600B": [
        "WU_E_DM_CONTENTCHANGED",
        "A download must be restarted because the update content changed in a new revision.",
    ],
    "0x80246FFF": [
        "WU_E_DM_UNEXPECTED",
        "There was a download manager error not covered by another WU_E_DM_* error code.",
    ],
    "0x80242000": [
        "WU_E_UH_REMOTEUNAVAILABLE",
        "A request for a remote update handler could not be completed because no remote process is available.",
    ],
    "0x80242001": [
        "WU_E_UH_LOCALONLY",
        "A request for a remote update handler could not be completed because the handler is local only.",
    ],
    "0x80242002": [
        "WU_E_UH_UNKNOWNHANDLER",
        "A request for an update handler could not be completed because the handler could not be recognized.",
    ],
    "0x80242003": [
        "WU_E_UH_REMOTEALREADYACTIVE",
        "A remote update handler could not be created because one already exists.",
    ],
    "0x80242004": [
        "WU_E_UH_DOESNOTSUPPORTACTION",
        "A request for the handler to install (uninstall) an update could not be completed because the update does not "
        "support install (uninstall).",
    ],
    "0x80242005": ["WU_E_UH_WRONGHANDLER", "An operation did not complete because the wrong handler was specified."],
    "0x80242006": [
        "WU_E_UH_INVALIDMETADATA",
        "A handler operation could not be completed because the update contains invalid metadata.",
    ],
    "0x80242007": [
        "WU_E_UH_INSTALLERHUNG",
        "An operation could not be completed because the installer exceeded the time limit.",
    ],
    "0x80242008": ["WU_E_UH_OPERATIONCANCELLED", "An operation being done by the update handler was cancelled."],
    "0x80242009": [
        "WU_E_UH_BADHANDLERXML",
        "An operation could not be completed because the handler-specific metadata is invalid.",
    ],
    "0x8024200A": [
        "WU_E_UH_CANREQUIREINPUT",
        "A request to the handler to install an update could not be completed because the update requires user input.",
    ],
    "0x8024200B": ["WU_E_UH_INSTALLERFAILURE", "The installer failed to install (uninstall) one or more updates."],
    "0x8024200C": [
        "WU_E_UH_FALLBACKTOSELFCONTAINED",
        "The update handler should download self-contained content rather than delta-compressed content for the "
        "update.",
    ],
    "0x8024200D": [
        "WU_E_UH_NEEDANOTHERDOWNLOAD",
        "The update handler did not install the update because it needs to be downloaded again.",
    ],
    "0x8024200E": [
        "WU_E_UH_NOTIFYFAILURE",
        "The update handler failed to send notification of the status of the install (uninstall) operation.",
    ],
    "0x8024200F": [
        "WU_E_UH_INCONSISTENT_FILE_NAMES",
        "The file names contained in the update metadata and in the update package are inconsistent.",
    ],
    "0x80242010": ["WU_E_UH_FALLBACKERROR", "The update handler failed to fall back to the self-contained content."],
    "0x80242011": [
        "WU_E_UH_TOOMANYDOWNLOADREQUESTS",
        "The update handler has exceeded the maximum number of download requests.",
    ],
    "0x80242012": ["WU_E_UH_UNEXPECTEDCBSRESPONSE", "The update handler has received an unexpected response from CBS."],
    "0x80242013": ["WU_E_UH_BADCBSPACKAGEID", "The update metadata contains an invalid CBS package identifier."],
    "0x80242014": ["WU_E_UH_POSTREBOOTSTILLPENDING", "The post-reboot operation for the update is still in progress."],
    "0x80242015": [
        "WU_E_UH_POSTREBOOTRESULTUNKNOWN",
        "The result of the post-reboot operation for the update could not be determined.",
    ],
    "0x80242016": [
        "WU_E_UH_POSTREBOOTUNEXPECTEDSTATE",
        "The state of the update after its post-reboot operation has completed is unexpected.",
    ],
    "0x80242017": [
        "WU_E_UH_NEW_SERVICING_STACK_REQUIRED",
        "The operating system servicing stack must be updated before this update is downloaded or installed.",
    ],
    "0x80242FFF": ["WU_E_UH_UNEXPECTED", "An update handler error not covered by another WU_E_UH_* code."],
    "0x80248000": ["WU_E_DS_SHUTDOWN", "An operation failed because Windows Update Agent is shutting down."],
    "0x80248001": ["WU_E_DS_INUSE", "An operation failed because the data store was in use."],
    "0x80248002": ["WU_E_DS_INVALID", "The current and expected states of the data store do not match."],
    "0x80248003": ["WU_E_DS_TABLEMISSING", "The data store is missing a table."],
    "0x80248004": ["WU_E_DS_TABLEINCORRECT", "The data store contains a table with unexpected columns."],
    "0x80248005": [
        "WU_E_DS_INVALIDTABLENAME",
        "A table could not be opened because the table is not in the data store.",
    ],
    "0x80248006": ["WU_E_DS_BADVERSION", "The current and expected versions of the data store do not match."],
    "0x80248007": ["WU_E_DS_NODATA", "The information requested is not in the data store."],
    "0x80248008": [
        "WU_E_DS_MISSINGDATA",
        "The data store is missing required information or has a NULL in a table column that requires a non-null "
        "value.",
    ],
    "0x80248009": [
        "WU_E_DS_MISSINGREF",
        "The data store is missing required information or has a reference to missing license terms, file, localized "
        "property or linked row.",
    ],
    "0x8024800A": [
        "WU_E_DS_UNKNOWNHANDLER",
        "The update was not processed because its update handler could not be recognized.",
    ],
    "0x8024800B": [
        "WU_E_DS_CANTDELETE",
        "The update was not deleted because it is still referenced by one or more services.",
    ],
    "0x8024800C": [
        "WU_E_DS_LOCKTIMEOUTEXPIRED",
        "The data store section could not be locked within the allotted time.",
    ],
    "0x8024800D": [
        "WU_E_DS_NOCATEGORIES",
        "The category was not added because it contains no parent categories and is not a top-level category itself.",
    ],
    "0x8024800E": ["WU_E_DS_ROWEXISTS", "The row was not added because an existing row has the same primary key."],
    "0x8024800F": [
        "WU_E_DS_STOREFILELOCKED",
        "The data store could not be initialized because it was locked by another process.",
    ],
    "0x80248010": [
        "WU_E_DS_CANNOTREGISTER",
        "The data store is not allowed to be registered with COM in the current process.",
    ],
    "0x80248011": ["WU_E_DS_UNABLETOSTART", "Could not create a data store object in another process."],
    "0x80248013": [
        "WU_E_DS_DUPLICATEUPDATEID",
        "The server sent the same update to the client with two different revision IDs.",
    ],
    "0x80248014": [
        "WU_E_DS_UNKNOWNSERVICE",
        "An operation did not complete because the service is not in the data store.",
    ],
    "0x80248015": [
        "WU_E_DS_SERVICEEXPIRED",
        "An operation did not complete because the registration of the service has expired.",
    ],
    "0x80248016": [
        "WU_E_DS_DECLINENOTALLOWED",
        "A request to hide an update was declined because it is a mandatory update or because it was deployed with a "
        "deadline.",
    ],
    "0x80248017": [
        "WU_E_DS_TABLESESSIONMISMATCH",
        "A table was not closed because it is not associated with the session.",
    ],
    "0x80248018": [
        "WU_E_DS_SESSIONLOCKMISMATCH",
        "A table was not closed because it is not associated with the session.",
    ],
    "0x80248019": [
        "WU_E_DS_NEEDWINDOWSSERVICE",
        "A request to remove the Windows Update service or to unregister it with Automatic Updates was declined "
        "because it is a built-in service and/or Automatic Updates cannot fall back to another service.",
    ],
    "0x8024801A": ["WU_E_DS_INVALIDOPERATION", "A request was declined because the operation is not allowed."],
    "0x8024801B": [
        "WU_E_DS_SCHEMAMISMATCH",
        "The schema of the current data store and the schema of a table in a backup XML document do not match.",
    ],
    "0x8024801C": [
        "WU_E_DS_RESETREQUIRED",
        "The data store requires a session reset; release the session and retry with a new session.",
    ],
    "0x8024801D": [
        "WU_E_DS_IMPERSONATED",
        "A data store operation did not complete because it was requested with an impersonated identity.",
    ],
    "0x80248FFF": ["WU_E_DS_UNEXPECTED", "A data store error not covered by another WU_E_DS_* code."],
    "0x8024C001": ["WU_E_DRV_PRUNED", "A driver was skipped."],
    "0x8024C002": [
        "WU_E_DRV_NOPROP_OR_LEGACY",
        "A property for the driver could not be found. It may not conform with required specifications.",
    ],
    "0x8024C003": ["WU_E_DRV_REG_MISMATCH", "The registry type read for the driver does not match the expected type."],
    "0x8024C004": ["WU_E_DRV_NO_METADATA", "The driver update is missing metadata."],
    "0x8024C005": ["WU_E_DRV_MISSING_ATTRIBUTE", "The driver update is missing a required attribute."],
    "0x8024C006": ["WU_E_DRV_SYNC_FAILED", "Driver synchronization failed."],
    "0x8024C007": [
        "WU_E_DRV_NO_PRINTER_CONTENT",
        "Information required for the synchronization of applicable printers is missing.",
    ],
    "0x8024CFFF": ["WU_E_DRV_UNEXPECTED", "A driver error not covered by another WU_E_DRV_* code."],
    "0x80240001": ["WU_E_NO_SERVICE", "Windows Update Agent was unable to provide the service."],
    "0x80240002": ["WU_E_MAX_CAPACITY_REACHED", "The maximum capacity of the service was exceeded."],
    "0x80240003": ["WU_E_UNKNOWN_ID", "An ID cannot be found."],
    "0x80240004": ["WU_E_NOT_INITIALIZED", "The object could not be initialized."],
    "0x80240005": [
        "WU_E_RANGEOVERLAP",
        "The update handler requested a byte range overlapping a previously requested range.",
    ],
    "0x80240006": ["WU_E_TOOMANYRANGES", "The requested number of byte ranges exceeds the maximum number (2^31 - 1)."],
    "0x80240007": ["WU_E_INVALIDINDEX", "The index to a collection was invalid."],
    "0x80240008": ["WU_E_ITEMNOTFOUND", "The key for the item queried could not be found."],
    "0x80240009": [
        "WU_E_OPERATIONINPROGRESS",
        "Another conflicting operation was in progress. Some operations such as installation cannot be performed twice "
        "simultaneously.",
    ],
    "0x8024000A": ["WU_E_COULDNOTCANCEL", "Cancellation of the operation was not allowed."],
    "0x8024000B": ["WU_E_CALL_CANCELLED", "Operation was cancelled."],
    "0x8024000C": ["WU_E_NOOP", "No operation was required."],
    "0x8024000D": [
        "WU_E_XML_MISSINGDATA",
        "Windows Update Agent could not find required information in the update's XML data.",
    ],
    "0x8024000E": ["WU_E_XML_INVALID", "Windows Update Agent found invalid information in the update's XML data."],
    "0x8024000F": ["WU_E_CYCLE_DETECTED", "Circular update relationships were detected in the metadata."],
    "0x80240010": ["WU_E_TOO_DEEP_RELATION", "Update relationships too deep to evaluate were evaluated."],
    "0x80240011": ["WU_E_INVALID_RELATIONSHIP", "An invalid update relationship was detected."],
    "0x80240012": ["WU_E_REG_VALUE_INVALID", "An invalid registry value was read."],
    "0x80240013": ["WU_E_DUPLICATE_ITEM", "Operation tried to add a duplicate item to a list."],
    "0x80240016": [
        "WU_E_INSTALL_NOT_ALLOWED",
        "Operation tried to install while another installation was in progress or the system was pending a mandatory "
        "restart.",
    ],
    "0x80240017": ["WU_E_NOT_APPLICABLE", "Operation was not performed because there are no applicable updates."],
    "0x80240018": ["WU_E_NO_USERTOKEN", "Operation failed because a required user token is missing."],
    "0x80240019": [
        "WU_E_EXCLUSIVE_INSTALL_CONFLICT",
        "An exclusive update cannot be installed with other updates at the same time.",
    ],
    "0x8024001A": ["WU_E_POLICY_NOT_SET", "A policy value was not set."],
    "0x8024001B": [
        "WU_E_SELFUPDATE_IN_PROGRESS",
        "The operation could not be performed because the Windows Update Agent is self-updating.",
    ],
    "0x8024001D": ["WU_E_INVALID_UPDATE", "An update contains invalid metadata."],
    "0x8024001E": [
        "WU_E_SERVICE_STOP",
        "Operation did not complete because the service or system was being shut down.",
    ],
    "0x8024001F": ["WU_E_NO_CONNECTION", "Operation did not complete because the network connection was unavailable."],
    "0x80240020": [
        "WU_E_NO_INTERACTIVE_USER",
        "Operation did not complete because there is no logged-on interactive user.",
    ],
    "0x80240021": ["WU_E_TIME_OUT", "Operation did not complete because it timed out."],
    "0x80240022": ["WU_E_ALL_UPDATES_FAILED", "Operation failed for all the updates."],
    "0x80240023": ["WU_E_EULAS_DECLINED", "The license terms for all updates were declined."],
    "0x80240024": ["WU_E_NO_UPDATE", "There are no updates."],
    "0x80240025": ["WU_E_USER_ACCESS_DISABLED", "Group Policy settings prevented access to Windows Update."],
    "0x80240026": ["WU_E_INVALID_UPDATE_TYPE", "The type of update is invalid."],
    "0x80240027": ["WU_E_URL_TOO_LONG", "The URL exceeded the maximum length."],
    "0x80240028": [
        "WU_E_UNINSTALL_NOT_ALLOWED",
        "The update could not be uninstalled because the request did not originate from a WSUS server.",
    ],
    "0x80240029": [
        "WU_E_INVALID_PRODUCT_LICENSE",
        "Search may have missed some updates before there is an unlicensed application on the system.",
    ],
    "0x8024002A": ["WU_E_MISSING_HANDLER", "A component required to detect applicable updates was missing."],
    "0x8024002B": ["WU_E_LEGACYSERVER", "An operation did not complete because it requires a newer version of server."],
    "0x8024002C": [
        "WU_E_BIN_SOURCE_ABSENT",
        "A delta-compressed update could not be installed because it required the source.",
    ],
    "0x8024002D": ["WU_E_SOURCE_ABSENT", "A full-file update could not be installed because it required the source."],
    "0x8024002E": ["WU_E_WU_DISABLED", "Access to an unmanaged server is not allowed."],
    "0x8024002F": [
        "WU_E_CALL_CANCELLED_BY_POLICY",
        "Operation did not complete because the DisableWindowsUpdateAccess policy was set.",
    ],
    "0x80240030": ["WU_E_INVALID_PROXY_SERVER", "The format of the proxy list was invalid."],
    "0x80240031": ["WU_E_INVALID_FILE", "The file is in the wrong format."],
    "0x80240032": ["WU_E_INVALID_CRITERIA", "The search criteria string was invalid."],
    "0x80240033": ["WU_E_EULA_UNAVAILABLE", "License terms could not be downloaded."],
    "0x80240034": ["WU_E_DOWNLOAD_FAILED", "Update failed to download."],
    "0x80240035": ["WU_E_UPDATE_NOT_PROCESSED", "The update was not processed."],
    "0x80240036": ["WU_E_INVALID_OPERATION", "The object's current state did not allow the operation."],
    "0x80240037": ["WU_E_NOT_SUPPORTED", "The functionality for the operation is not supported."],
    "0x80240038": ["WU_E_WINHTTP_INVALID_FILE", "The downloaded file has an unexpected content type."],
    "0x80240039": ["WU_E_TOO_MANY_RESYNC", "Agent is asked by server to resync too many times."],
    "0x80240040": ["WU_E_NO_SERVER_CORE_SUPPORT", "WUA API method does not run on Server Core installation."],
    "0x80240041": ["WU_E_SYSPREP_IN_PROGRESS", "Service is not available while sysprep is running."],
    "0x80240042": ["WU_E_UNKNOWN_SERVICE", "The update service is no longer registered with AU."],
    "0x80240043": ["WU_E_NO_UI_SUPPORT", "There's no support for WUA UI."],
    "0x80240FFF": ["WU_E_UNEXPECTED", "An operation failed due to reasons not covered by another error code."],
    "0x80070422": ["NA", "This issue occurs when the Windows Update service stops working or isn't running."],
    "0x00240001": ["WU_S_SERVICE_STOP", "Windows Update Agent was stopped successfully."],
    "0x00240002": ["WU_S_SELFUPDATE", "Windows Update Agent updated itself."],
    "0x00240003": ["WU_S_UPDATE_ERROR", "Operation completed successfully but there were errors applying the updates."],
    "0x00240004": [
        "WU_S_MARKED_FOR_DISCONNECT",
        "A callback was marked to be disconnected later because the request to disconnect the operation came while a "
        "callback was executing.",
    ],
    "0x00240005": ["WU_S_REBOOT_REQUIRED", "The system must be restarted to complete installation of the update."],
    "0x00240006": ["WU_S_ALREADY_INSTALLED", "The update to be installed is already installed on the system."],
    "0x00240007": ["WU_S_ALREADY_UNINSTALLED", "The update to be removed isn't installed on the system."],
    "0x00240008": ["WU_S_ALREADY_DOWNLOADED", "The update to be downloaded has already been downloaded."],
    "0x80241001": [
        "WU_E_MSI_WRONG_VERSION",
        "Search may have missed some updates because the Windows Installer is less than version 3.1.",
    ],
    "0x80241002": [
        "WU_E_MSI_NOT_CONFIGURED",
        "Search may have missed some updates because the Windows Installer is not configured.",
    ],
    "0x80241003": [
        "WU_E_MSP_DISABLED",
        "Search may have missed some updates because policy has disabled Windows Installer patching.",
    ],
    "0x80241004": [
        "WU_E_MSI_WRONG_APP_CONTEXT",
        "An update could not be applied because the application is installed per-user.",
    ],
    "0x80241FFF": [
        "WU_E_MSP_UNEXPECTED",
        "Search may have missed some updates because there was a failure of the Windows Installer.",
    ],
    "0x8024D001": [
        "WU_E_SETUP_INVALID_INFDATA",
        "Windows Update Agent could not be updated because an INF file contains invalid information.",
    ],
    "0x8024D002": [
        "WU_E_SETUP_INVALID_IDENTDATA",
        "Windows Update Agent could not be updated because the wuident.cab file contains invalid information.",
    ],
    "0x8024D003": [
        "WU_E_SETUP_ALREADY_INITIALIZED",
        "Windows Update Agent could not be updated because of an internal error that caused setup initialization to be "
        "performed twice.",
    ],
    "0x8024D004": [
        "WU_E_SETUP_NOT_INITIALIZED",
        "Windows Update Agent could not be updated because setup initialization never completed successfully.",
    ],
    "0x8024D005": [
        "WU_E_SETUP_SOURCE_VERSION_MISMATCH",
        "Windows Update Agent could not be updated because the versions specified in the INF do not match the actual "
        "source file versions.",
    ],
    "0x8024D006": [
        "WU_E_SETUP_TARGET_VERSION_GREATER",
        "Windows Update Agent could not be updated because a WUA file on the target system is newer than the "
        "corresponding source file.",
    ],
    "0x8024D007": [
        "WU_E_SETUP_REGISTRATION_FAILED",
        "Windows Update Agent could not be updated because regsvr32.exe returned an error.",
    ],
    "0x8024D009": [
        "WU_E_SETUP_SKIP_UPDATE",
        "An update to the Windows Update Agent was skipped due to a directive in the wuident.cab file.",
    ],
    "0x8024D00A": [
        "WU_E_SETUP_UNSUPPORTED_CONFIGURATION",
        "Windows Update Agent could not be updated because the current system configuration is not supported.",
    ],
    "0x8024D00B": [
        "WU_E_SETUP_BLOCKED_CONFIGURATION",
        "Windows Update Agent could not be updated because the system is configured to block the update.",
    ],
    "0x8024D00C": [
        "WU_E_SETUP_REBOOT_TO_FIX",
        "Windows Update Agent could not be updated because a restart of the system is required.",
    ],
    "0x8024D00D": ["WU_E_SETUP_ALREADYRUNNING", "Windows Update Agent setup is already running."],
    "0x8024D00E": [
        "WU_E_SETUP_REBOOTREQUIRED",
        "Windows Update Agent setup package requires a reboot to complete installation.",
    ],
    "0x8024D00F": [
        "WU_E_SETUP_HANDLER_EXEC_FAILURE",
        "Windows Update Agent could not be updated because the setup handler failed during execution.",
    ],
    "0x8024D010": [
        "WU_E_SETUP_INVALID_REGISTRY_DATA",
        "Windows Update Agent could not be updated because the registry contains invalid information.",
    ],
    "0x8024D013": [
        "WU_E_SETUP_WRONG_SERVER_VERSION",
        "Windows Update Agent could not be updated because the server does not contain update information for this "
        "version.",
    ],
    "0x8024DFFF": [
        "WU_E_SETUP_UNEXPECTED",
        "Windows Update Agent could not be updated because of an error not covered by another WU_E_SETUP_* error code.",
    ],
    "0x80070BC9": [
        "ERROR_FAIL_REBOOT_REQUIRED",
        "The requested operation failed. Restart the system to roll back changes made.",
    ],
    "0x80200053": ["BG_E_VALIDATION_FAILED", "NA"],
    "0x80072EFD": ["TIME_OUT_ERRORS", "The operation timed out"],
    "0x80072EFE": [
        "WININET_E_CONNECTION_ABORTED; The connection with the server was closed abnormally",
        "BITS is unable to transfer the file successfully.",
    ],
    "0x80D02002": ["TIME_OUT_ERRORS", "The operation timed out"],
    "0X8007000D": ["ERROR_INVALID_DATA", "Indicates data that isn't valid was downloaded or corruption occurred."],
    "0x8024A10A": ["USO_E_SERVICE_SHUTTING_DOWN", "Indicates that the Windows Update Service is shutting down."],
    "0x80246017": [
        "WU_E_DM_UNAUTHORIZED_LOCAL_USER",
        "The download failed because the local user was denied authorization to download the content.",
    ],
    "0x800f0821": [
        "CBS_E_ABORT; client abort, IDABORT returned by ICbsUIHandler method except Error()",
        "CBS transaction timeout exceeded.",
    ],
    "0x800f0825": [
        "CBS_E_CANNOT_UNINSTALL; Package can't be uninstalled.",
        "Typically this error is due component store corruption caused when a component is in a partially installed "
        "state.",
    ],
    "0x800F0920": [
        "CBS_E_HANG_DETECTED; A failure to respond was detected while processing the operation.",
        "Subsequent error logged after getting 0x800f0821",
    ],
    "0x800f081f": [
        "CBS_E_SOURCE_MISSING; source for package or file not found, ResolveSource() unsuccessful",
        "Component Store corruption",
    ],
    "0x800f0831": ["CBS_E_STORE_CORRUPTION; CBS store is corrupted.", "Corruption in the Windows Component Store."],
    "0x80070005": [
        "E_ACCESSDENIED; General access denied error",
        "File system or registry key permissions have been changed and the servicing stack doesn't have the required "
        "level of access.",
    ],
    "0x80070003": [
        "ERROR_PATH_NOT_FOUND; The system can't find the path specified.",
        "The servicing stack can't access a specific path.",
    ],
    "0x80073701": [
        "ERROR_SXS_ASSEMBLY_MISSING; The referenced assembly couldn't be found.",
        "Typically, a component store corruption caused when a component is in a partially installed state.",
    ],
    "0x80072F8F": [
        "WININET_E_DECODING_FAILED; Content decoding has failed",
        "TLS 1.2 isn't configured correctly on the client.",
    ],
    "0x80072EE2": [
        "WININET_E_TIMEOUT; The operation timed out",
        "Unable to scan for updates due to a connectivity issue to Windows Update, Configuration Manager, or WSUS.",
    ],
    "0x80070490": ["ERROR_NOT_FOUND", "This error occurs during driver installation as part of the update."],
    "0x800f0922": [
        "CBS_E_INSTALLERS_FAILED",
        "The July cumulative update failed to be installed on Windows Server 2016",
    ],
    "0x80070bc9": [
        "ERROR_FAIL_REBOOT_REQUIRED",
        "The TrustedInstaller service startup type is set to 'Manual' by Group Policy (GPO), which prevented it from "
        "starting to complete pending operations.",
    ],
    "0x800706be": [
        "Failed to install cumulative updates",
        "Windows Server 2016 Std failed to install cumulative packages by using the .msu package. No error is "
        "returned. When installing the packages with dism.exe, it returned the error 0x800706be.",
    ],
}


class WuaHistoryPlugin(Plugin):
    """Plugin to return all available historical Windows Update Agent operations stored in the DataStore.edb."""

    DATASTORE_PATH = "%windir%/softwaredistribution/datastore/datastore.edb"
    DATASTORE_UPDATE_TABLE = "tbHistory"

    def __init__(self, target: Target):
        super().__init__(target)

        self._datastore = None
        self._update_table = None

        if (path := target.resolve(self.DATASTORE_PATH)).exists():
            try:
                self._datastore = EseDB(path.open())
                self._update_table = self._datastore.table(self.DATASTORE_UPDATE_TABLE)
            except Exception as e:
                self.target.log.warning("Error opening Windows Update Agent datastore")
                self.target.log.debug("", exc_info=e)

    def check_compatible(self) -> None:
        if not self._datastore or not self._update_table:
            raise UnsupportedPluginError("No Windows Update Agent data found.")

    def get_table_records(self) -> Iterator[dict[str, str | int | datetime]]:
        for record in self._update_table.records():
            record_data = {}
            for column in self._update_table.columns:
                column_value = record.get(column.name)

                if mapped_column_name := TBHISTORY_COLUMN_MAP.get(column.name):
                    if mapped_column_name in "unmapped_result":
                        column_value = hex(int(column_value) & 0xFFFFFFFF)

                        info = WUA_CODE_MAP.get(column_value, ["Unknown", "Unknown"])
                        record_data[f"{mapped_column_name}_description"] = info[0]
                        record_data[f"{mapped_column_name}_string"] = info[1]

                    record_data.update(self._format_record_value(mapped_column_name, column_value))

            yield record_data

    @export(record=WuaHistoryRecord)
    def wua_history(self) -> Iterator[WuaHistoryRecord]:
        """Returns all available historical Windows Update Agent operations stored in the DataStore.edb.

        The Windows Update Agent (WUA) stores information of its operations in a DataStore.edb file.
        Historical data of these operations, successful or failed performed patches, are stored in the 'tbHistory' table
        of this database. The plugin extracts all rows of this table. For certain columns of the table the extracted
        data is mapped to publicly available enumerations to provide more meaningful information.

        With this plugin you should be able to ascertain the patch level of systems, and it may aid you into
        finding why a system was generating a lot of other types of events in a certain time period.

        References:
            - https://learn.microsoft.com/en-us/windows/deployment/update/how-windows-update-works
            - https://learn.microsoft.com/en-us/windows/win32/api/wuapi/nn-wuapi-iupdatehistoryentry
            - https://learn.microsoft.com/en-us/windows/deployment/update/windows-update-error-reference
            - https://learn.microsoft.com/en-us/troubleshoot/windows-client/installing-updates-features-roles/common-windows-update-errors
            - https://learn.microsoft.com/en-us/troubleshoot/windows-client/installing-updates-features-roles/common-windows-update-errors?context=%2Fwindows%2Fdeployment%2Fcontext%2Fcontext
            - https://github.com/libyal/esedb-kb/blob/main/documentation/Windows%20Update.asciidoc
            - https://www.nirsoft.net/articles/extract-windows-updates-list-external-drive.html

        Yields a WuaHistoryRecord with fields:

        .. code-block:: text

            ts (datetime): The timestamp (UTC) of when the Windows Update Agent operation was finished.
            categories (string): Category of the update.
            classification (string): Unique ID indicating which classification the update has.
            classification_mapped (string): Mapping of the 'classification' field, giving an understandable classification.
            client_id (string): Client that initiated the Windows Update Agent operation.
            description (string): Description of the update.
            flags (int): Undocumented and unknown.
            id_event (int): Index number of the Windows Update Agent record in the tbHistory table.
            kb (string): Another unique ID of the update.
            status_id (int): Integer signifying result of operation.

            status_mapped (string): Mapping of the 'status' field.
            server_selection (int): The update service that was used for the Windows Update Agent operation.
            server_selection_mapped (string): Mapping of the 'server_selection' field.
            title (string): Title of the update.
            mapped_result (string): The mapped result code of an update operation.
            mapped_result_string (string): Mapping of the 'mapped_result' field, giving the error string.
            mapped_result_description (string): Mapping of the 'mapped_result' field, giving a description of the error.
            unmapped_result (string): The unmapped result code of an update operation.
            unmapped_result_string (string): Mapping of the 'unmapped_result' field, giving the error string.
            unmapped_result_description (string): Mapping of the 'unmapped_result' field, giving a description of the error.
            update_id (string): Unique ID of the performed update.
            server_id (string): Unique ID of the service used for the update operation.
            server_id_mapped (string): Mapping of the 'server_id' field, indication the service used for the update operation.
            support_url (string): Support URL of the update.
            uninstall_notes (string): Uninstall notes of the update.
            uninstall_steps (string): Uninstall steps of the update.
            more_info_url (string): Additional informational URLs of the update.
            path (uri): Path of the datastore containing the Windows Update Agent records.
            id_user (int): Undocumented and unknown.
            is_service_is_additional (string): Undocumented and unknown.
        """  # noqa: E501
        for record in self.get_table_records():
            values = {TBHISTORY_COLUMN_MAP.get(key, key): value for key, value in record.items()}

            yield WuaHistoryRecord(
                **values,
                path=self.target.fs.path(self.DATASTORE_PATH),
                _target=self.target,
            )

    def _format_record_value(self, mapped_column_name: str, value: str) -> dict[str, str | int | datetime]:
        format_data = {}

        if mapped_column_name == "ts":
            format_data[mapped_column_name] = oatimestamp(value)
        elif mapped_column_name == "update_id":
            format_data[mapped_column_name] = decode_guid(value[:16])
        elif mapped_column_name == "server_id":
            guid = decode_guid(value[:16])
            format_data[mapped_column_name] = guid
            format_data["server_id_mapped"] = SERVER_ID_MAP.get(guid, "Unknown")
        elif mapped_column_name == "classification":
            format_data[mapped_column_name] = value
            format_data["classification_mapped"] = CLASSIFICATION_MAP.get(value, "Unknown")
        elif mapped_column_name == "title":
            format_data[mapped_column_name] = value
            if kb := re.search(r"(KB.[0-9]*)", value):
                format_data["kb"] = kb.group()
        elif mapped_column_name == "status_id":
            format_data[mapped_column_name] = value
            format_data["status_mapped"] = STATUS_MAP.get(value, "Unknown")
        elif mapped_column_name == "server_selection":
            key = value - 1
            format_data[mapped_column_name] = key
            format_data["server_selection_mapped"] = SERVER_SELECTION_MAP.get(key, "Unknown")
        else:
            format_data[mapped_column_name] = value

        return format_data



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.child.parallels import ParallelsChildTargetPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_parallels_child_detection(target_macos_users: Target, fs_macos: VirtualFilesystem) -> None:
    """Test if we correctly find Parallels child VMs on MacOS targets."""

    fs_macos.makedirs("Users/dissect/Parallels/Windows 11.pvm")
    fs_macos.makedirs("Users/dissect/Documents/Parallels/Windows 10.pvm")
    fs_macos.makedirs(
        "Users/dissect/Library/Group Containers/someversionnumber.com.parallels.desktop.appstore/Shared/Parallels/Windows 8.pvm"  # noqa: E501
    )
    fs_macos.makedirs("Users/Shared/Parallels/Windows 7.pvm")

    target_macos_users.add_plugin(ParallelsChildTargetPlugin)
    children = list(target_macos_users.list_children())

    assert len(children) == 4
    assert [c.path for c in children] == [
        "/Users/Shared/Parallels/Windows 7.pvm",
        "/Users/dissect/Parallels/Windows 11.pvm",
        "/Users/dissect/Documents/Parallels/Windows 10.pvm",
        "/Users/dissect/Library/Group Containers/someversionnumber.com.parallels.desktop.appstore/Shared/Parallels/Windows 8.pvm",  # noqa: E501
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, Callable
from unittest.mock import patch

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.pvm import PvmLoader
from dissect.target.target import Target
from tests._utils import mkdirs

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path


@pytest.fixture
def mock_pvm_dir(tmp_path: Path) -> Iterator[Path]:
    mkdirs(tmp_path, ["Test.pvm"])
    (tmp_path / "Test.pvm" / "config.pvs").touch()

    with patch("dissect.hypervisor.descriptor.pvs.PVS") as mock_pvs:
        mock_pvs.return_value = mock_pvs
        mock_pvs.disks.return_value = ["mock.hdd"]

        yield tmp_path / "Test.pvm"


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], mock_pvm_dir: Path) -> None:
    """Test that we correctly use ``PvmLoader`` when opening a ``Target``."""
    with patch("dissect.target.container.open"), patch("dissect.target.target.Target.apply"):
        target = opener(mock_pvm_dir)
        assert isinstance(target._loader, PvmLoader)
        assert target.path == mock_pvm_dir


def test_loader(mock_pvm_dir: Path) -> None:
    """Test that ``PvmLoader`` correctly loads a PVM file and its disks."""
    with patch("dissect.target.container.open") as mock_container_open:
        loader = loader_open(mock_pvm_dir)
        assert isinstance(loader, PvmLoader)

        t = Target()
        loader.map(t)

        assert len(t.disks) == 1
        mock_container_open.assert_called_with(mock_pvm_dir.resolve() / "mock.hdd")



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import LoaderError
from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.loaders.ovf import OvfLoader

if TYPE_CHECKING:
    from pathlib import Path


class OvaLoader(OvfLoader):
    """Load Open Virtual Appliance (OVA) files.

    References:
        - https://en.wikipedia.org/wiki/Open_Virtualization_Format
    """

    def __init__(self, path: Path, **kwargs):
        self.ova = TarFilesystem(path.open("rb"))

        ovf_path = next(self.ova.path().glob("*.ovf"), None)
        if ovf_path is None:
            raise LoaderError("Invalid OVA file (can't find .ovf)")

        super().__init__(ovf_path, **kwargs)

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() == ".ova"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target import filesystem, volume
from dissect.target.containers.vhdx import VhdxContainer
from dissect.target.loader import Loader
from dissect.target.loaders.dir import find_and_map_dirs, find_dirs
from dissect.target.plugin import OperatingSystem

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


# KAPE doesn't have the correct filenames for several files, like $J or $Secure_$SDS
# The same applies to Velociraptor offline collector Windows.KapeFiles.Targets
USNJRNL_PATHS = ["$Extend/$J", "$Extend/$UsnJrnl$J"]


def open_vhdx(path: Path) -> Iterator[Filesystem]:
    container = VhdxContainer(path)
    volume_system = volume.open(container)
    for vol in volume_system.volumes:
        yield filesystem.open(vol)


def is_valid_kape_dir(path: Path) -> bool:
    os_type, dirs = find_dirs(path)
    if os_type == OperatingSystem.WINDOWS:
        for dir_path in dirs:
            for path in USNJRNL_PATHS:
                if dir_path.joinpath(path).exists():
                    return True

    return False


def is_valid_kape_vhdx(path: Path) -> bool:
    if path.suffix == ".vhdx":
        try:
            for fs in open_vhdx(path):
                return is_valid_kape_dir(fs.path())
        except Exception:
            return False

    return False


class KapeLoader(Loader):
    """Load KAPE forensic image format files.

    References:
        - https://www.kroll.com/en/insights/publications/cyber/kroll-artifact-parser-extractor-kape
    """

    @staticmethod
    def detect(path: Path) -> bool:
        if path.is_dir():
            return is_valid_kape_dir(path)
        if path.suffix.lower() == ".vhdx":
            return is_valid_kape_vhdx(path)
        return False

    def map(self, target: Target) -> None:
        path = self.absolute_path if self.absolute_path.is_dir() else next(open_vhdx(self.absolute_path)).path()

        find_and_map_dirs(
            target,
            path,
            sds_path="$Secure_$SDS",
            usnjrnl_path=USNJRNL_PATHS,
        )



from datetime import datetime, timezone

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.windows.everything._plugin import EverythingPlugin
from dissect.target.target import Target
from tests._utils import absolute_path

DIRECTORY = "directory"
FILE = "file"


def test_everything_1_7_20_ntfs_only(target_win: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_NTFS_ONLY.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = list(target_win.everything.locate())
    assert len(records) == 126828

    recycle_bin = records[0]
    assert recycle_bin.type == DIRECTORY
    assert recycle_bin.path == "C:\\$Recycle.Bin"
    assert recycle_bin.date_modified == datetime(2023, 12, 16, 9, 8, 42, 726189, tzinfo=timezone.utc)

    entry = records[-40000]
    assert entry.type == FILE
    assert entry.path == "C:\\Windows\\System32\\msidntld.dll"

    assert entry.date_modified == datetime(2021, 5, 8, 8, 14, 34, 642525, tzinfo=timezone.utc)


def test_everything_1_7_20_ntfs_and_folder(target_win: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_NTFS_AND_FOLDER.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = list(target_win.everything.locate())
    assert len(records) == 127042
    recycle_bin_c = records[0]
    assert recycle_bin_c.type == DIRECTORY
    assert recycle_bin_c.path == "C:\\$Recycle.Bin"
    assert recycle_bin_c.date_modified == datetime(2023, 12, 16, 9, 8, 42, 726189, tzinfo=timezone.utc)

    recycle_bin_e = records[1]
    assert recycle_bin_e.type == DIRECTORY
    assert recycle_bin_e.path == "E:\\$RECYCLE.BIN"
    assert recycle_bin_e.date_modified == datetime(2024, 1, 27, 9, 54, 32, tzinfo=timezone.utc)

    entry = records[94783]
    assert entry.type == FILE
    assert entry.path == "E:\\potato.txt"
    assert entry.date_modified == datetime(2024, 1, 27, 9, 54, 32, tzinfo=timezone.utc)


def test_everything_1_7_20_ntfs_and_refs(target_win: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_NTFS_AND_REFS.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = list(target_win.everything.locate())
    assert len(records) == 127227
    recycle_bin_c = records[0]
    assert recycle_bin_c.type == DIRECTORY
    assert recycle_bin_c.path == "C:\\$Recycle.Bin"
    assert recycle_bin_c.date_modified == datetime(2023, 12, 16, 9, 8, 42, 726189, tzinfo=timezone.utc)

    recycle_bin_e = records[1]
    assert recycle_bin_e.type == DIRECTORY
    assert recycle_bin_e.path == "E:\\$RECYCLE.BIN"
    assert recycle_bin_e.date_modified == datetime(2024, 1, 27, 21, 37, 6, 675776, tzinfo=timezone.utc)

    entry = records[94937]
    assert entry.type == FILE
    assert entry.path == "E:\\potato.txt"
    assert entry.date_modified == datetime(2024, 1, 27, 21, 37, 13, 425810, tzinfo=timezone.utc)

    entry = records[108011]
    assert entry.type == FILE
    assert entry.path == "E:\\test.txt"
    assert entry.date_modified == datetime(2024, 1, 27, 21, 56, 57, 597593, tzinfo=timezone.utc)


def test_everything_1_7_20_refs_with_disabled_ntfs(target_win: Target, fs_win: VirtualFilesystem) -> None:
    # Just making sure that disabling a drive doesn't mess up parsing
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_REFS_WITH_DISABLED_NTFS.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = list(target_win.everything.locate())
    assert len(records) == 7
    recycle_bin_e = records[0]
    assert recycle_bin_e.type == DIRECTORY
    assert recycle_bin_e.path == "E:\\$RECYCLE.BIN"
    assert recycle_bin_e.date_modified == datetime(2024, 1, 27, 21, 37, 6, 675776, tzinfo=timezone.utc)

    entry = records[5]
    assert entry.type == FILE
    assert entry.path == "E:\\potato.txt"
    assert entry.date_modified == datetime(2024, 1, 27, 21, 37, 13, 425810, tzinfo=timezone.utc)


def test_everything_1_7_20_refs_include_only(target_win: Target, fs_win: VirtualFilesystem) -> None:
    # Just making sure that including specific subfolders doesn't mess up parsing
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_REFS_INCLUDE_ONLY.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = list(target_win.everything.locate())
    assert len(records) == 4
    recycle_bin_e = records[0]
    assert recycle_bin_e.type == DIRECTORY
    assert recycle_bin_e.path == "E:\\test"
    assert recycle_bin_e.date_modified == datetime(2024, 1, 27, 22, 6, 9, 597820, tzinfo=timezone.utc)

    entry = records[3]
    assert entry.type == FILE
    assert entry.path == "E:\\test\\test.txt"
    assert entry.date_modified == datetime(2024, 1, 27, 21, 56, 57, 597593, tzinfo=timezone.utc)


def test_everything_1_7_20_efu(target_win: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_FILE_LIST.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = list(target_win.everything.locate())
    assert len(records) == 5
    entry = records[0]
    assert entry.type == DIRECTORY
    assert entry.path == "E:"
    assert entry.date_modified is None

    entry = records[1]
    assert entry.type == DIRECTORY
    assert entry.path == "E:\\test"
    assert entry.date_modified == datetime(2024, 1, 27, 22, 6, 9, 597820, tzinfo=timezone.utc)


def test_everything_1_7_17_ntfs_refs_efu_folder(target_win: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "\\Program Files\\Everything\\Everything.db",
        absolute_path("_data/plugins/os/windows/everything/1.7.17/Everything_NTFS_REFS_EFU_FOLDER.db.gz"),
        compression="gzip",
    )
    target_win.add_plugin(EverythingPlugin)

    records = list(target_win.everything.locate())
    assert len(records) == 127577
    entry = records[0]
    assert entry.type == DIRECTORY
    assert entry.path == "C:\\$Recycle.Bin"
    assert entry.date_modified == datetime(2023, 12, 16, 9, 8, 42, 726189, tzinfo=timezone.utc)
    assert entry.date_created is None

    entry = records[1]
    assert entry.type == DIRECTORY
    assert entry.path == "C:\\$WinREAgent"
    assert entry.date_modified == datetime(2024, 1, 26, 20, 49, 55, 41714, tzinfo=timezone.utc)


@pytest.mark.parametrize(
    "map_path",
    [
        "\\Program Files\\Everything\\Everything.db",
        "\\Program Files\\Everything\\Everything.COMPNAME.USERNAME.db",
        "\\Program Files (x86)\\Everything\\Everything.db",
        "\\Users\\John\\AppData\\Local\\Everything\\Everything.db",
        "\\Users\\John\\AppData\\Local\\Everything\\Everything.COMPNAME.John.db",
    ],
)
def test_everything_path_mapper(target_win_users: Target, fs_win: VirtualFilesystem, map_path: str) -> None:
    fs_win.map_file(
        map_path,
        absolute_path("_data/plugins/os/windows/everything/1.7.20/Everything_NTFS_ONLY.db.gz"),
        compression="gzip",
    )
    assert EverythingPlugin(target_win_users).check_compatible() is None



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.apps.browser import brave
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.fixture
def target_brave(target_win_users: Target, fs_win: VirtualFilesystem) -> Target:
    base_path_default = "Users\\John\\AppData\\Local\\BraveSoftware\\Brave-Browser\\User Data\\Default"
    base_path_profile = "Users\\John\\AppData\\Local\\BraveSoftware\\Brave-Browser\\User Data\\Profile 1"
    files = [
        ("History", "_data/plugins/apps/browser/chrome/generic/History"),
        ("Preferences", "_data/plugins/apps/browser/chrome/generic/Preferences"),
        ("Secure Preferences", "_data/plugins/apps/browser/chrome/generic/Secure Preferences"),
        ("Network\\Cookies", "_data/plugins/apps/browser/chromium/Cookies"),
        ("Login Data", "_data/plugins/apps/browser/chromium/Login Data"),
    ]

    for filename, test_path in files:
        fs_win.map_file(f"{base_path_default}\\{filename}", absolute_path(test_path))
        fs_win.map_file(f"{base_path_profile}\\{filename}", absolute_path(test_path))

    target_win_users.add_plugin(brave.BravePlugin)

    return target_win_users


def test_brave_history(target_brave: Target) -> None:
    records = list(target_brave.brave.history())
    assert len(records) == 10


def test_brave_downloads(target_brave: Target) -> None:
    records = list(target_brave.brave.downloads())
    assert len(records) == 2

    assert records[0].ts_start == datetime(2023, 2, 24, 11, 54, 19, 726147, tzinfo=timezone.utc)
    assert records[0].ts_end == datetime(2023, 2, 24, 11, 54, 21, 30043, tzinfo=timezone.utc)
    assert records[0].id == 6
    assert records[0].path == "C:\\Users\\John\\Downloads\\dissect-main (2).zip"
    assert records[0].url == "https://codeload.github.com/fox-it/dissect/zip/refs/heads/main"
    assert records[0].size == 17924
    assert records[0].state == "complete"


def test_brave_extensions(target_brave: Target) -> None:
    records = list(target_brave.brave.extensions())
    assert len(records) == 16


def test_brave_cookies(target_brave: Target) -> None:
    records = list(target_brave.brave.cookies())
    assert len(records) == 10
    assert all(record.host == ".tweakers.net" for record in records)


def test_windows_edge_passwords_plugin(target_brave: Target) -> None:
    records = list(target_brave.brave.passwords())

    assert len(records) == 4

    for record in records:
        assert record.browser == "brave"
        assert record.decrypted_username == "username"
        assert record.decrypted_password is None

    assert records[0].url == "https://example.com/"
    assert records[1].url == "https://example.org/"



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

import pytest

from dissect.target.exceptions import FileNotFoundError
from dissect.target.filesystems.config import (
    ConfigurationEntry,
    ConfigurationFilesystem,
)
from dissect.target.plugins.general.config import ConfigurationTreePlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.fixture
def config_tree(target_unix: Target) -> ConfigurationTreePlugin:
    target_unix.add_plugin(ConfigurationTreePlugin)
    return target_unix.config_tree


def test_config_tree_plugin(config_tree: ConfigurationTreePlugin, fs_unix: VirtualFilesystem, tmp_path: Path) -> None:
    tmp_path.joinpath("new/path").mkdir(parents=True, exist_ok=True)
    tmp_path.joinpath("new/config").mkdir(parents=True, exist_ok=True)
    fs_unix.map_dir("/etc", tmp_path)
    fs_unix.map_file("/etc/new/path/config", absolute_path("_data/helpers/configutil/config"))

    options = {"separator": (r"\s",)}

    assert isinstance(config_tree("/etc/new/path/config"), ConfigurationEntry)
    assert isinstance(config_tree("/etc/new/path/config", **options).get("help"), ConfigurationEntry)
    assert isinstance(config_tree("/etc/new/path/config/help", **options), ConfigurationEntry)
    assert isinstance(config_tree.get(), ConfigurationFilesystem)
    assert isinstance(config_tree.get("/etc/new/path/config/help", **options), ConfigurationEntry)

    assert sorted(config_tree("/etc/new/path/config", as_dict=True, **options).keys()) == [
        "help",
        "test",
    ]
    assert isinstance(config_tree.get("/etc/new/path/config/help", as_dict=True, **options), list)

    with pytest.raises(FileNotFoundError):
        config_tree("/etc/new/path/help", **options)


@pytest.mark.parametrize(
    "collapse_type",
    [
        tuple,
        dict,
        set,
        list,
    ],
)
def test_collapse_types(
    config_tree: ConfigurationTreePlugin, fs_unix: VirtualFilesystem, collapse_type: type[Iterable]
) -> None:
    """Using specifically the SequenceType due to using lru_cache in the plugin."""

    fs_unix.map_file_fh("/etc/new/path/config", BytesIO(b"key=value"))

    config_tree("/etc/new/path/config", collapse=collapse_type())


@pytest.mark.parametrize(
    ("hint", "data_bytes"),
    [
        ("ini", b"[DEFAULT]\nkey=value"),
        ("xml", b"<a>currently_just_text</a>"),
        ("json", b'{"key": "value"}'),
        ("yaml", b"key: value"),
        ("cnf", b"key=value"),
        ("conf", b"key value"),
        ("sample", b"currently_just_text"),
        ("template", b"currently_just_text"),
        ("toml", b"key = 'value'"),
    ],
)
def test_as_dict(
    config_tree: ConfigurationTreePlugin, fs_unix: VirtualFilesystem, hint: str, data_bytes: bytes
) -> None:
    fs_unix.map_file_fh("/etc/new/path/config", BytesIO(data_bytes))
    assert isinstance(config_tree("/etc/new/path/config", hint=hint, as_dict=True), dict)



from __future__ import annotations

import binascii
from io import BytesIO
from typing import TYPE_CHECKING, Any, Callable
from unittest.mock import Mock, mock_open, patch

import pytest
from dissect.util.ts import wintimestamp

from dissect.target.plugins.os.windows.regf.shimcache import (
    MAGIC_NT52,
    MAGIC_NT61,
    MAGIC_WIN10,
    MAGIC_WIN81,
    SHIMCACHE_WIN_TYPE,
    TYPE_VARIATIONS,
    CRCMismatchException,
    ShimCache,
    ShimcachePlugin,
    c_shim,
)

if TYPE_CHECKING:
    import datetime
    from collections.abc import Iterator

    from dissect.target.target import Target

TEST_ARGUMENTS = {
    SHIMCACHE_WIN_TYPE.VERSION_WIN81_NO_HEADER: {"path_len": 1},
    SHIMCACHE_WIN_TYPE.VERSION_WIN81: {"ts": 0, "pkg_len": 0, "pkg": "", "path_len": 0},
}


def create_header_bytes(magic: int, position: int = 0, length: int = 0x100) -> BytesIO:
    if position > length - 4:
        raise ValueError("total header bytes would be larger than length.")

    result_array = b"\x00" * length
    output_array = result_array[0:position] + magic.to_bytes(4, "little") + result_array[position + 4 :]
    return BytesIO(output_array)


@pytest.fixture
def mocked_shimcache() -> ShimCache:
    with patch.object(ShimCache, "identify"):
        return ShimCache(fh=BytesIO(), ntversion="", noheader=False)


@pytest.fixture(params=[True, False])
def crc_check(request: pytest.FixtureRequest) -> bool:
    return request.param


@pytest.fixture
def mocked_nt61_file(version: int, pathname: str) -> tuple[dict, Mock]:
    winnt_variations = TYPE_VARIATIONS.get(version)

    nt_header = winnt_variations.get("header")
    nt_header_data = nt_header(magic=0, num_entries=1).dumps()
    nt_entry_header = winnt_variations.get("header_function")
    nt_entry_header_data = nt_entry_header(nt_header_data)().dumps()

    open_file = mock_open().return_value
    open_file.read.side_effect = [
        nt_header_data,
        nt_entry_header_data,
        pathname,
    ]

    return winnt_variations, open_file


@pytest.fixture
def created_header_win8plus(crc_check: bool, version: int) -> tuple[dict, Mock]:
    variations = TYPE_VARIATIONS.get(version)
    entry_header, data_header = variations.get("headers")

    additional_kwargs = TEST_ARGUMENTS.get(version, {})

    data = data_header(path="", **additional_kwargs).dumps()
    entry = entry_header()

    if hasattr(entry, "data"):
        entry.len = len(data)
        entry.data = data

    if crc_check:
        entry.crc = binascii.crc32(data)

    mocked_file = mock_open(read_data=entry.dumps())
    return variations, mocked_file


@pytest.fixture(params=[0, 1])
def offset(request: pytest.FixtureRequest) -> int:
    return request.param


@pytest.fixture
def mocked_nt52_file(offset: int, pathname: str) -> tuple[dict, Mock]:
    winnt_variations = TYPE_VARIATIONS.get(SHIMCACHE_WIN_TYPE.VERSION_NT52)

    nt_header = winnt_variations.get("header")
    nt_header_data = nt_header(magic=0, num_entries=1).dumps()
    nt_entry_header = winnt_variations.get("header_function")
    nt_entry_header_data = nt_entry_header(b"\x00" * 32)(_padding=offset).dumps()

    length = len(c_shim.NT52_ENTRY_32 if offset else c_shim.NT52_ENTRY_64)

    open_file = mock_open().return_value
    open_file.read.side_effect = [
        nt_header_data,
        nt_entry_header_data[: len(c_shim.NT52_ENTRY_32)],
        nt_entry_header_data[:length],
        pathname,
    ]

    return winnt_variations, open_file


@pytest.mark.parametrize("shimcache_error", [NotImplementedError, EOFError])
def test_shimcache_plugin_initialize(target_win: Target, shimcache_error: type[Exception]) -> None:
    shimcache = ShimcachePlugin(target_win)
    mocked_registry_keys = Mock()
    target_win.registry.keys = Mock(return_value=[mocked_registry_keys])
    mocked_registry_keys.value.return_value.value = b"hello_world"

    with patch(ShimCache.__module__, side_effect=[shimcache_error]):
        assert list(shimcache.shimcache()) == []


@pytest.mark.parametrize(
    ("file_data", "ntversion", "expected_value"),
    [
        (create_header_bytes(MAGIC_NT52), "6.3", SHIMCACHE_WIN_TYPE.VERSION_NT52),
        (create_header_bytes(MAGIC_NT61), "6.3", SHIMCACHE_WIN_TYPE.VERSION_NT61),
        (create_header_bytes(MAGIC_WIN81), "6.3", SHIMCACHE_WIN_TYPE.VERSION_WIN81_NO_HEADER),
        (create_header_bytes(MAGIC_WIN81, 0x80, 0x84), "6.3", SHIMCACHE_WIN_TYPE.VERSION_WIN81),
        (create_header_bytes(MAGIC_WIN10, 0x30, 0x34), "6.3", SHIMCACHE_WIN_TYPE.VERSION_WIN10),
        (create_header_bytes(MAGIC_WIN10, 0x34, 0x38), "6.3", SHIMCACHE_WIN_TYPE.VERSION_WIN10_CREATORS),
    ],
)
def test_shimcache_identify(mocked_shimcache: ShimCache, file_data: bytes, ntversion: str, expected_value: int) -> None:
    mocked_shimcache.fh = file_data
    mocked_shimcache.ntversion = ntversion
    assert mocked_shimcache.identify() == expected_value


@pytest.mark.parametrize(
    ("file_data", "ntversion", "no_header", "expected_value"),
    [
        (create_header_bytes(0), "6.3", False, SHIMCACHE_WIN_TYPE.VERSION_WIN81),
        (create_header_bytes(0), "6.3", True, SHIMCACHE_WIN_TYPE.VERSION_WIN81_NO_HEADER),
        (create_header_bytes(0), None, False, NotImplementedError),
        (create_header_bytes(MAGIC_WIN81), None, False, NotImplementedError),
        (create_header_bytes(0, 0x80, 0x84), None, False, NotImplementedError),
        (create_header_bytes(0, 0x30, 0x34), None, False, NotImplementedError),
        (create_header_bytes(0, 0x34, 0x38), None, False, NotImplementedError),
    ],
)
def test_identify_special_conditions(
    mocked_shimcache: ShimCache,
    file_data: bytes,
    ntversion: str | None,
    no_header: bool,
    expected_value: int | type[Exception],
) -> None:
    mocked_shimcache.fh = file_data
    mocked_shimcache.ntversion = ntversion
    mocked_shimcache.noheader = no_header

    if ntversion:
        assert mocked_shimcache.identify() == expected_value
    else:
        with pytest.raises(expected_value):
            mocked_shimcache.identify()


@pytest.mark.parametrize(
    ("version", "expected_method"),
    [
        (SHIMCACHE_WIN_TYPE.VERSION_WIN10_CREATORS, ShimCache.iter_win_8_plus),
        (SHIMCACHE_WIN_TYPE.VERSION_WIN10, ShimCache.iter_win_8_plus),
        (SHIMCACHE_WIN_TYPE.VERSION_WIN81, ShimCache.iter_win_8_plus),
        (SHIMCACHE_WIN_TYPE.VERSION_NT61, ShimCache.iter_nt),
        (SHIMCACHE_WIN_TYPE.VERSION_NT52, ShimCache.iter_nt),
    ],
)
def test_shim_cache_iterator_mocked(mocked_shimcache: ShimCache, version: int, expected_method: Callable) -> None:
    mocked_shimcache.version = version
    mocked_shimcache.ntversion = "6.3"

    with patch.object(ShimCache, expected_method.__name__) as iterator_method:
        assert iter(mocked_shimcache) == iterator_method.return_value


def test_shim_cache_iterator_unimplemented(mocked_shimcache: ShimCache) -> None:
    mocked_shimcache.version = -1
    with pytest.raises(NotImplementedError):
        iter(mocked_shimcache)


@pytest.mark.parametrize(
    ("version", "timestamp"),
    [
        (SHIMCACHE_WIN_TYPE.VERSION_WIN10, wintimestamp(0)),
        (SHIMCACHE_WIN_TYPE.VERSION_WIN10_CREATORS, wintimestamp(0)),
        (SHIMCACHE_WIN_TYPE.VERSION_WIN81_NO_HEADER, None),
        (SHIMCACHE_WIN_TYPE.VERSION_WIN81, wintimestamp(0)),
    ],
)
def test_shim_iter_win8_plus(
    mocked_shimcache: ShimCache, created_header_win8plus: tuple[dict, Mock], timestamp: datetime.datetime
) -> None:
    win_10_variations, mocked_file = created_header_win8plus

    mocked_shimcache.fh = mocked_file.return_value

    output = next(iter(mocked_shimcache.iter_win_8_plus(**win_10_variations)))
    if isinstance(output, tuple):
        assert output == (timestamp, "")
    else:
        assert isinstance(output, CRCMismatchException)


@pytest.mark.parametrize(
    ("version", "pathname", "expected_output"),
    [
        (
            SHIMCACHE_WIN_TYPE.VERSION_NT61,
            b"h\x00e\x00l\x00l\x00o\x00_\x00w\x00o\x00r\x00l\x00d\x00",
            [(wintimestamp(0), "hello_world")],
        ),
        (SHIMCACHE_WIN_TYPE.VERSION_NT61, b"hello_world", []),
    ],
)
def test_shim_iter_nt61(
    mocked_shimcache: ShimCache, mocked_nt61_file: tuple[dict, Mock], expected_output: list
) -> None:
    winnt_variations, open_file = mocked_nt61_file
    mocked_shimcache.fh = open_file

    assert list(mocked_shimcache.iter_nt(**winnt_variations)) == expected_output


@pytest.mark.parametrize(
    ("pathname", "expected_output"),
    [
        (
            b"h\x00e\x00l\x00l\x00o\x00_\x00w\x00o\x00r\x00l\x00d\x00",
            [(wintimestamp(0), "hello_world")],
        ),
        (b"hello_world", []),
    ],
)
def test_shim_iter_nt52(
    mocked_shimcache: ShimCache, mocked_nt52_file: tuple[dict, Mock], expected_output: list
) -> None:
    winnt_variations, open_file = mocked_nt52_file
    mocked_shimcache.fh = open_file

    assert list(mocked_shimcache.iter_nt(**winnt_variations)) == expected_output


def list_generator(input_list: Iterator[Any]) -> Iterator[Any]:
    yield from input_list


def test_gracefull_shutdown_crc(target_win: Target, mocked_shimcache: ShimCache) -> None:
    plugin = ShimcachePlugin(target_win)
    mocked_shimcache.version = SHIMCACHE_WIN_TYPE.VERSION_WIN10

    with patch(f"{ShimcachePlugin.__module__}.ShimcacheRecord") as mocked_record:
        list_gen = list_generator([CRCMismatchException(), (0, "path")])
        assert list(plugin._get_records(None, list_gen)) == [mocked_record.return_value]



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import ClassVar

from dissect.target.helpers.sunrpc.sunrpc import ProcedureDescriptor

# See https://datatracker.ietf.org/doc/html/rfc1057


NfsProgram = 100003
MountProgram = 100005
NfsVersion = 3

# List of procedure descriptors
MountProc = ProcedureDescriptor(MountProgram, NfsVersion, 1)
ReadDirPlusProc = ProcedureDescriptor(NfsProgram, NfsVersion, 17)
ReadFileProc = ProcedureDescriptor(NfsProgram, NfsVersion, 6)
LookupProc = ProcedureDescriptor(NfsProgram, NfsVersion, 3)
GetAttrProc = ProcedureDescriptor(NfsProgram, NfsVersion, 1)
ReadLinkProc = ProcedureDescriptor(NfsProgram, NfsVersion, 5)


class NfsStat(IntEnum):
    OK = 0
    ERR_PERM = 1
    ERR_NOENT = 2
    ERR_IO = 5
    ERR_NXIO = 6
    ERR_ACCES = 13
    ERR_EXIST = 17
    ERR_XDEV = 18
    ERR_NODEV = 19
    ERR_NOTDIR = 20
    ERR_ISDIR = 21
    ERR_INVAL = 22
    ERR_FBIG = 27
    ERR_NOSPC = 28
    ERR_ROFS = 30
    ERR_MLINK = 31
    ERR_NAMETOOLONG = 63
    ERR_NOTEMPTY = 66
    ERR_DQUOT = 69
    ERR_STALE = 70
    ERR_REMOTE = 71
    ERR_BADHANDLE = 10001


@dataclass
class FileHandle:
    MAXSIZE: ClassVar[int] = 64
    opaque: bytes

    def __post_init__(self):
        if len(self.opaque) > self.MAXSIZE:
            raise ValueError(f"FileHandle3 cannot exceed {self.MAXSIZE} bytes")


@dataclass
class CookieVerf:
    MAXSIZE: ClassVar[int] = 8
    opaque: bytes

    def __post_init__(self):
        if len(self.opaque) > self.MAXSIZE:
            raise ValueError(f"CookieVerf cannot exceed {self.MAXSIZE} bytes")


class FileType(IntEnum):
    REG = 1  # regular file
    DIR = 2  # directory
    BLK = 3  # block special
    CHR = 4  # character special
    LNK = 5  # symbolic link
    SOCK = 6  # socket
    FIFO = 7  # fifo


@dataclass
class SpecData:
    specdata1: int
    specdata2: int


@dataclass
class NfsTime:
    seconds: int
    nseconds: int


@dataclass
class FileAttributes:
    type: FileType
    mode: int
    nlink: int
    uid: int
    gid: int
    size: int
    used: int
    rdev: SpecData
    fsid: int
    fileid: int
    atime: NfsTime
    mtime: NfsTime
    ctime: NfsTime


class MountStat(IntEnum):
    OK = 0  # no error
    ERR_PERM = 1  # Not owner
    ERR_NOENT = 2  # No such file or directory
    ERR_IO = 5  # I/O error
    ERR_ACCES = 13  # Permission denied
    ERR_NOTDIR = 20  # Not a directory
    ERR_INVAL = 22  # Invalid argument
    ERR_NAMETOOLONG = 63  # Filename too long
    ERR_NOTSUPP = 10004  # Operation not supported
    ERR_SERVERFAULT = 10006  # A failure on the server


@dataclass
class MountOK:
    filehandle: FileHandle
    auth_flavors: list[int]


@dataclass
class ReadDirPlusParams:
    dir: FileHandle
    cookie: int
    cookieverf: CookieVerf
    dir_count: int
    max_count: int


@dataclass
class EntryPlus:
    fileid: int
    name: str
    cookie: int
    attributes: FileAttributes | None
    handle: FileHandle | None


@dataclass
class ReadDirPlusResult:  # READDIRPLUS3res in RFC
    dir_attributes: FileAttributes | None
    cookieverf: CookieVerf
    entries: list[EntryPlus]
    eof: bool


@dataclass
class ReadParams:  # READ3args in RFC
    file: FileHandle
    offset: int
    count: int


@dataclass
class ReadResult:  # READ3resok in RFC
    file_attributes: FileAttributes | None
    count: int
    eof: bool
    data: bytes


@dataclass
class DirOpArgs:
    handle: FileHandle
    filename: str


@dataclass
class LookupResult:  # LOOKUP3resok in RFC
    object: FileHandle
    obj_attributes: FileAttributes | None
    dir_attributes: FileAttributes | None


@dataclass
class ReadlinkResult:  # READLINK3resok in RFC
    obj_attributes: FileAttributes | None
    target: str  # named "data" in the RFC



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.dir import DirLoader, find_dirs
from dissect.target.plugin import OperatingSystem
from dissect.target.target import Target
from tests._utils import mkdirs

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], tmp_path: Path) -> None:
    """Test that we correctly use ``DirLoader`` when opening a ``Target``."""
    root = tmp_path
    mkdirs(root, ["windows/system32"])

    target = opener(root)
    assert isinstance(target._loader, DirLoader)
    assert target.path == root


def test_windows(tmp_path: Path) -> None:
    """Test the ``DirLoader`` for Windows directories."""
    root = tmp_path
    mkdirs(root, ["windows/system32"])

    os_type, dirs = find_dirs(root)
    assert os_type == OperatingSystem.WINDOWS
    assert len(dirs) == 1

    loader = loader_open(root)
    assert isinstance(loader, DirLoader)

    t = Target()
    loader.map(t)
    assert len(t.filesystems) == 1


def test_winnt(tmp_path: Path) -> None:
    """Test the ``DirLoader`` for WinNT directories."""
    root = tmp_path
    mkdirs(tmp_path, ["winnt"])

    os_type, dirs = find_dirs(root)
    assert os_type == OperatingSystem.WINDOWS
    assert len(dirs) == 1

    loader = loader_open(root)
    assert isinstance(loader, DirLoader)

    t = Target()
    loader.map(t)
    assert len(t.filesystems) == 1


def test_windows_drive_letters(tmp_path: Path) -> None:
    """Test the ``DirLoader`` with Windows drive letters."""
    root = tmp_path
    mkdirs(root, ["C/windows/system32", "D/test", "E/test"])

    os_type, dirs = find_dirs(root)
    assert os_type == OperatingSystem.WINDOWS
    assert len(dirs) == 3

    loader = loader_open(root)
    assert isinstance(loader, DirLoader)

    t = Target()
    loader.map(t)
    assert len(t.filesystems) == 3
    assert len(t.fs.mounts) == 3


def test_linux(tmp_path: Path) -> None:
    """Test the ``DirLoader`` for Linux directories."""
    root = tmp_path
    mkdirs(root, ["etc", "var"])

    os_type, dirs = find_dirs(root)
    assert os_type == OperatingSystem.LINUX
    assert len(dirs) == 1

    loader = loader_open(root)
    assert isinstance(loader, DirLoader)

    t = Target()
    loader.map(t)
    assert len(t.filesystems) == 1


def test_macos(tmp_path: Path) -> None:
    """Test the ``DirLoader`` for macOS directories."""
    root = tmp_path
    mkdirs(root, ["Library"])

    os_type, dirs = find_dirs(root)
    assert os_type == OperatingSystem.OSX
    assert len(dirs) == 1

    loader = loader_open(root)
    assert isinstance(loader, DirLoader)

    t = Target()
    loader.map(t)
    assert len(t.filesystems) == 1



from __future__ import annotations

import errno
import os
import sys
import traceback
from typing import Callable


class Error(Exception):
    """Generic dissect.target error."""

    def __init__(self, message: str | None = None, extra: list[Exception] | None = None):
        if extra:
            exceptions = "\n\n".join(["".join(traceback.format_exception_only(type(e), e)) for e in extra])
            message = f"{message}\n\nAdditionally, the following exceptions occurred:\n\n{exceptions}"

        super().__init__(message)
        self.__extra__ = extra


# Use FatalError if you don't want your error to be buried by other errors
# but just mark it as fatal so the tools 'on top' may proceed to shutdown
class FatalError(Error):
    """An error occurred that cannot be resolved."""

    def emit_last_message(self, emitter: Callable) -> None:
        emitter(str(self))
        os.dup2(os.open(os.devnull, os.O_RDWR), sys.stdout.fileno())
        os.dup2(os.open(os.devnull, os.O_RDWR), sys.stderr.fileno())


class TargetError(Error):
    """A target error occurred."""


class LoaderError(Error):
    """A loader error occurred."""


class PluginError(Error):
    """A plugin error occurred."""


class ContainerError(Error):
    """A container error occurred."""


class VolumeSystemError(Error):
    """A volume system error occurred."""


class FilesystemError(Error):
    """A filesystem error occurred."""


class InvalidTaskError(Error):
    """A invalid XML file."""


class RegistryKeyNotFoundException(Error):
    """The registry was not found."""


class UnsupportedPluginError(PluginError):
    """The requested plugin is not supported by the target."""

    def root_cause_str(self) -> str:
        """Often with this type of Error, the root cause is more descriptive for the user."""
        return str(self.__cause__.args[0])


class PluginNotFoundError(PluginError):
    """Plugin cannot be found."""


class FileNotFoundError(FilesystemError, FileNotFoundError):
    """The requested path could not be found."""

    errno = errno.ENOENT


class IsADirectoryError(FilesystemError, IsADirectoryError):
    """The entry is a directory."""

    errno = errno.EISDIR


class NotADirectoryError(FilesystemError, NotADirectoryError):
    """The entry is not a directory."""

    errno = errno.ENOTDIR


class NotASymlinkError(FilesystemError):
    """The entry is not a symlink."""

    errno = errno.EINVAL


class SymlinkRecursionError(FilesystemError):
    """A symlink loop is detected for the entry."""

    errno = errno.ELOOP


class RegistryError(Error):
    """A registry error occurred."""


class RegistryKeyNotFoundError(RegistryError):
    """The requested registry key could not be found."""


class RegistryValueNotFoundError(RegistryError):
    """The requested registry value could not be found."""


class HiveUnavailableError(RegistryError):
    """The requested hive is unavailable."""


class RegistryCorruptError(RegistryError):
    """The registry is corrupt."""


class ConfigurationParsingError(Error):
    """An error occurred during configuration parsing."""


class TargetPathNotFoundError(TargetError):
    """The path to the target does not exist."""



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from dissect.target.plugins.os.windows import sru
from tests._utils import absolute_path

if TYPE_CHECKING:
    import pytest

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_sru_plugin(target_win: Target, fs_win: VirtualFilesystem, caplog: pytest.LogCaptureFixture) -> None:
    srudb = absolute_path("_data/plugins/os/windows/sru/SRUDB.dat")

    fs_win.map_file("Windows/System32/sru/SRUDB.dat", srudb)

    target_win.add_plugin(sru.SRUPlugin)

    assert len(list(target_win.sru())) == 220
    assert len(list(target_win.sru.application())) == 203
    assert len(list(target_win.sru.network_connectivity())) == 3
    assert len(list(target_win.sru.sdp_volume_provider())) == 6
    assert len(list(target_win.sru.sdp_physical_disk_provider())) == 3
    assert len(list(target_win.sru.sdp_cpu_provider())) == 3

    caplog.clear()
    with caplog.at_level(logging.WARNING, target_win.log.name):
        assert list(target_win.sru.vfu()) == []
        assert "Table not found: vfu" in caplog.text



from __future__ import annotations

import re
from collections import OrderedDict
from typing import TYPE_CHECKING, NamedTuple

from dissect.target.exceptions import RegistryError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import (
    TargetRecordDescriptor,
    create_extended_descriptor,
)
from dissect.target.plugin import OperatingSystem, Plugin, export, internal

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

EnvironmentRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/environment",
    [
        ("string", "name"),
        ("string", "value"),
    ],
)

PathextRecord = TargetRecordDescriptor(
    "windows/pathext",
    [
        ("string", "pathext"),
    ],
)


class EnvVarDetails(NamedTuple):
    name: str
    reg_keys: tuple[str, ...] = ()
    reg_value: str | None = None
    default: str | None = None


class EnvironmentVariablePlugin(Plugin):
    """Plugin that provides access to global environment variables.

    Mostly used internally.
    """

    # More information on the variables below can be found at:
    # https://renenyffenegger.ch/notes/Windows/development/environment-variables/index
    VARIABLES = (
        EnvVarDetails(
            # The value 'sysvol' is dissect specific as we map the system drive
            # to the sysvol drive name / directory in the root fs.
            name="%systemdrive%",
            default="sysvol",
        ),
        EnvVarDetails(
            "%windir%",
            ("HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion",),
            "SystemRoot",
        ),
        EnvVarDetails(
            # Note that the Directory value in
            # HKLM\SYSTEM\CurrentControlSet\Control\Windows, which is sometimes
            # mentioned, contains the literal %SystemRoot%.
            "%systemroot%",
            ("HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion",),
            "SystemRoot",
        ),
        EnvVarDetails(
            # Since Windows Vista this has the same value as %programdata%
            # Before that it was %systemdrive%\Documents and Settings\All Users.
            "%allusersprofile%",
            ("HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList",),
            "ProgramData",
        ),
        EnvVarDetails(
            "%programdata%",
            ("HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList",),
            "ProgramData",
        ),
        EnvVarDetails(
            "%public%",
            ("HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList",),
            "Public",
        ),
        EnvVarDetails(
            "%commonprogramfiles%",
            ("HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion",),
            "CommonFilesDir",
        ),
        EnvVarDetails(
            "%commonprogramfiles(x86)%",
            ("HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion",),
            "CommonFilesDir (x86)",
        ),
        EnvVarDetails(
            "%commonprogramw6432%",
            ("HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion",),
            "CommonW6432Dir",
        ),
        EnvVarDetails(
            "%programfiles%",
            ("HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion",),
            "ProgramFilesDir",
        ),
        EnvVarDetails(
            "%programfiles(x86)%",
            ("HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion",),
            "ProgramFilesDir (x86)",
        ),
        EnvVarDetails(
            "%programw6432%",
            ("HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion",),
            "ProgramW6432Dir",
        ),
        EnvVarDetails(
            "%temp%",
            (
                "HKLM\\SOFTWARE\\DefaultUserEnvironment",
                "HKLM\\SYSTEM\\CurrentControlSet\\Session Manager\\Environment",
            ),
            "TEMP",
        ),
        EnvVarDetails(
            "%tmp%",
            (
                "HKLM\\SOFTWARE\\DefaultUserEnvironment",
                "HKLM\\SYSTEM\\CurrentControlSet\\Session Manager\\Environment",
            ),
            "TMP",
        ),
        EnvVarDetails(
            # Note that %path% is a concatenation of the given reg_keys, not an
            # override by the last one available.
            "%path%",
            (
                "HKLM\\SOFTWARE\\DefaultUserEnvironment",
                "HKLM\\SYSTEM\\CurrentControlSet\\Session Manager\\Environment",
            ),
            "Path",
        ),
    )

    USER_VARIABLES = (
        EnvVarDetails(
            "%userprofile%",
            ("HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList\\{user_sid}",),
            "ProfileImagePath",
        ),
        EnvVarDetails(
            name="%appdata%",
            default="%userprofile%\\appdata\\roaming",
        ),
        EnvVarDetails(
            name="%localappdata%",
            default="%userprofile%\\appdata\\local",
        ),
        EnvVarDetails(
            "%onedrive%",
            ("HKU\\{user_sid}\\Environment",),
            "Onedrive",
        ),
        EnvVarDetails(
            # This overrides the 'global' %temp% variable
            "%temp%",
            (
                "HKLM\\SOFTWARE\\DefaultUserEnvironment",
                "HKLM\\SYSTEM\\CurrentControlSet\\Session Manager\\Environment",
                "HKU\\{user_sid}\\Environment",
            ),
            "TEMP",
        ),
        EnvVarDetails(
            # This overrides the 'global' %tmp% variable
            "%tmp%",
            (
                "HKLM\\SOFTWARE\\DefaultUserEnvironment",
                "HKLM\\SYSTEM\\CurrentControlSet\\Session Manager\\Environment",
                "HKU\\{user_sid}\\Environment",
            ),
            "TMP",
        ),
        EnvVarDetails(
            # Note that %path% is a concatenation of the given reg_keys, not an
            # override by the last one available.
            "%path%",
            (
                "HKLM\\SOFTWARE\\DefaultUserEnvironment",
                "HKLM\\SYSTEM\\CurrentControlSet\\Session Manager\\Environment",
                "HKU\\{user_sid}\\Environment",
            ),
            "Path",
        ),
    )

    def __init__(self, target: Target):
        super().__init__(target)

        self._env = None
        self._pathext = None

    @staticmethod
    def _expand_env(value: str, env_vars: OrderedDict[str, str]) -> str:
        """Replace, in order, all occurences of the keys from env_vars in path
        with the respective values from env_vars.
        """
        for var_name, var_value in env_vars.items():
            # lambda is used to prevent re.sub from doing substitution logic itself
            value = re.sub(re.escape(var_name), lambda _, var_value=var_value: var_value, value, flags=re.IGNORECASE)
        return value

    @classmethod
    def _expand_env_vars(cls, env_vars: OrderedDict[str, str]) -> OrderedDict[str, str]:
        """Replace, in order, all occurrences of the keys from env_vars in the
        other values of env_vars with the respective values of env_vars.

        This mimics Windows' immediate expansion and means the order in which
        the variables are defined in env_vars matters.
        See also: https://devblogs.microsoft.com/oldnewthing/20060823-00/?p=29993
        """
        expanded_env_vars = OrderedDict()

        for var_name, var_value in env_vars.items():
            var_value = cls._expand_env(var_value, expanded_env_vars)
            expanded_env_vars[var_name] = var_value

        return expanded_env_vars

    def _get_env_vars(self, env_var_details: list[EnvVarDetails]) -> OrderedDict[str, str]:
        """Get the environment variables defined in the env_var_details list.

        An OrderedDict is returned with the variable names as the keys and the
        variable values as the values. The values are NOT expanded with the
        values of variables earlier in the order.
        """
        env_vars = OrderedDict()

        for env_var_detail in env_var_details:
            var_value = env_var_detail.default

            if env_var_detail.reg_keys and env_var_detail.reg_value:
                # If there are multiple reg_keys, values from later ones override
                # those from earlier ones. Except for the %path% variable,
                # which is concatenated (with a ; separator) instead of
                # being overwritten.
                for reg_key in env_var_detail.reg_keys:
                    try:
                        reg_key_value = self.target.registry.value(reg_key, env_var_detail.reg_value).value
                    except RegistryError:  # noqa: PERF203
                        pass
                    else:
                        if env_var_detail.name == "%path%" and var_value is not None:
                            var_value = ";".join((var_value.rstrip(";"), reg_key_value))
                        else:
                            var_value = reg_key_value

            if not var_value:
                continue

            env_vars[env_var_detail.name] = var_value

        return env_vars

    def _get_system_env_vars(self) -> OrderedDict[str, str]:
        """Get the system environment variables

        An OrderedDict is returned with the variable names as the keys and the
        variable values as the values. The values ARE expanded with the values
        of variables earlier in the order.
        """
        if self._env is None:
            self._env = self._get_env_vars(self.VARIABLES)
            self._env = self._expand_env_vars(self._env)

        return self._env

    def _get_user_env_vars(self, user_sid: str | None = None) -> OrderedDict[str, str]:
        """Get the environment variables as seen by the user of the given user SID.

        If no user_sid is given, the function gives back the system environment
        variables.

        An OrderedDict is returned with the variable names as the keys and the
        variable values as the values. The values ARE expanded with the values
        of variables earlier in the order.
        """
        env_vars = self._get_system_env_vars()
        if user_sid is not None:
            env_vars = env_vars.copy()
            user_env_var_details = []

            for env_var_details in self.USER_VARIABLES:
                reg_keys = tuple(reg_key.format(user_sid=user_sid) for reg_key in env_var_details.reg_keys)
                env_var_details = env_var_details._replace(reg_keys=reg_keys)
                user_env_var_details.append(env_var_details)

            user_vars = self._get_env_vars(user_env_var_details)

            # To preserve the order in USER_VARIABLES, variables which are also
            # present in the system env_vars OrderedDict must be removed
            # before updating it with the user_vars.
            for user_var in user_vars:
                if user_var in env_vars:
                    del env_vars[user_var]
            env_vars.update(user_vars)
            env_vars = self._expand_env_vars(env_vars)

        return env_vars

    def check_compatible(self) -> None:
        if self.target.os != OperatingSystem.WINDOWS:
            raise UnsupportedPluginError("Target operating system is not Windows")

    @internal
    def expand_env(self, path: str, user_sid: str | None = None) -> str:
        env_vars = self._get_user_env_vars(user_sid)
        return self._expand_env(path, env_vars)

    @internal
    def user_env(self, user_sid: str | None = None) -> OrderedDict[str, str]:
        """Return a dict of all found (user) environment variables.

        If no ``user_sid`` is provided, this function will return just the system environment variables.
        """
        return self._get_user_env_vars(user_sid)

    @property
    @internal
    def env(self) -> OrderedDict[str, str]:
        """Return a dict of all found system environment variables."""
        return self._get_system_env_vars()

    @export(record=EnvironmentRecord)
    def environment_variables(self) -> Iterator[EnvironmentRecord]:
        """Return all environment variables on a Windows system.

        Environment variables are dynamic-named values that can affect the way processes are running on the system.
        Examples variables are PATH, HOME and TEMP. Adversaries may alter or create environment variables to exploit
        a system.

        References:
            - https://en.wikipedia.org/wiki/Environment_variable
            - https://www.elttam.com/blog/env/
        """
        for name, value in self._get_system_env_vars().items():
            yield EnvironmentRecord(
                name=name,
                value=value,
                _target=self.target,
            )

        for user in self.target.users():
            for name, value in self._get_user_env_vars(user.sid).items():
                yield EnvironmentRecord(
                    name=name,
                    value=value,
                    _target=self.target,
                    _user=user,
                )

    def _get_pathext(self) -> set[str]:
        if self._pathext is None:
            self._pathext = set()

            try:
                env_key = "HKLM\\SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment"
                for key in self.target.registry.keys(env_key):
                    for ext in key.value("PATHEXT").value.lower().split(";"):
                        self._pathext.add(ext)
            except RegistryError:
                pass
        return self._pathext

    @property
    @internal
    def pathext(self) -> set[str]:
        """Return a list of all found path extensions."""
        return self._get_pathext()

    @export(record=PathextRecord)
    def path_extensions(self) -> Iterator[PathextRecord]:
        """Return all found path extensions."""
        for pathext in self._get_pathext():
            yield PathextRecord(
                pathext=pathext,
                _target=self.target,
            )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers import keychain
from dissect.target.plugins.os.windows.dpapi.dpapi import DPAPIPlugin
from dissect.target.plugins.os.windows.dpapi.keyprovider.keychain import KeychainKeyProviderPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_dpapi_keyprovider_keychain(target_win: Target) -> None:
    """test if we yield provided keychain items correctly."""
    keychain.register_key(
        key_type=keychain.KeyType.PASSPHRASE,
        value="password1",
        identifier=None,
        provider="user",
    )
    keychain.register_key(
        key_type=keychain.KeyType.PASSPHRASE,
        value="password2",
        identifier=None,
        provider=None,
    )

    target_win.add_plugin(DPAPIPlugin, check_compatible=False)
    target_win.add_plugin(KeychainKeyProviderPlugin)

    keys = list(target_win.dpapi.keyprovider.keychain())

    assert keys == [
        ("dpapi.keyprovider.keychain", "password1"),
        ("dpapi.keyprovider.keychain", "password2"),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

VirtualMachineRecord = TargetRecordDescriptor(
    "proxmox/vm",
    [
        ("string", "path"),
    ],
)


class VirtualMachinePlugin(Plugin):
    """Plugin to list Proxmox virtual machines."""

    def check_compatible(self) -> None:
        if self.target.os != "proxmox":
            raise UnsupportedPluginError("Not a Proxmox operating system")

    @export(record=VirtualMachineRecord)
    def vmlist(self) -> Iterator[VirtualMachineRecord]:
        """List Proxmox virtual machines on this node."""
        for config in self.target.fs.path("/etc/pve/qemu-server").iterdir():
            yield VirtualMachineRecord(
                path=config,
                _target=self.target,
            )



from __future__ import annotations

import os
import stat
from datetime import datetime, timezone
from typing import TYPE_CHECKING, TextIO

from dissect.target.exceptions import FileNotFoundError
from dissect.target.filesystem import FilesystemEntry, LayerFilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Sequence

    from dissect.target.helpers.fsutil import TargetPath

# ['mode', 'addr', 'dev', 'nlink', 'uid', 'gid', 'size', 'atime', 'mtime', 'ctime']
STAT_TEMPLATE = """  File: {path} {symlink}
  Size: {size}       Blocks: {blocks}    IO Block: {blksize}     {filetype}
Device: {device}     Inode: {inode}      Links: {nlink}
Access: ({modeord}/{modestr})  Uid: ( {uid} )   Gid: ( {gid} )
Access: {atime}
Modify: {mtime}
Change: {ctime}
 Birth: {btime}"""

FALLBACK_LS_COLORS = "rs=0:di=01;34:ln=01;36:mh=00:pi=40;33:so=01;35:do=01;35:bd=40;33;01:cd=40;33;01:or=40;31;01:mi=00:su=37;41:sg=30;43:ca=30;41:tw=30;42:ow=34;42:st=37;44:ex=01;32"  # noqa: E501


def prepare_ls_colors() -> dict[str, str]:
    """Parse the LS_COLORS environment variable so we can use it later."""
    d = {}
    ls_colors = os.environ.get("LS_COLORS", FALLBACK_LS_COLORS)
    for line in ls_colors.split(":"):
        if not line:
            continue

        ft, _, value = line.partition("=")
        ft = ft.removeprefix("*")

        d[ft] = f"\x1b[{value}m{{}}\x1b[0m"

    return d


LS_COLORS = prepare_ls_colors()


def fmt_ls_colors(ft: str, name: str) -> str:
    """Helper method to colorize strings according to LS_COLORS."""
    try:
        return LS_COLORS[ft].format(name)
    except KeyError:
        pass

    try:
        return LS_COLORS[fsutil.splitext(name)[1]].format(name)
    except KeyError:
        pass

    return name


def human_size(bytes: int, units: Sequence[str] = ("", "K", "M", "G", "T", "P", "E")) -> str:
    """Helper function to return the human readable string representation of bytes."""
    return str(bytes) + units[0] if bytes < 1024 else human_size(bytes >> 10, units[1:])


def stat_modestr(st: fsutil.stat_result) -> str:
    """Helper method for generating a mode string from a numerical mode value."""
    return stat.filemode(st.st_mode)


def print_extensive_file_stat_listing(
    stdout: TextIO,
    name: str,
    entry: FilesystemEntry | None = None,
    timestamp: datetime | None = None,
    human_readable: bool = False,
) -> None:
    """Print the file status as a single line."""
    if entry is not None:
        try:
            entry_stat = entry.lstat()
        except FileNotFoundError:
            pass
        else:
            if timestamp is None:
                timestamp = entry_stat.st_mtime
            symlink = f" -> {entry.readlink()}" if entry.is_symlink() else ""
            utc_time = datetime.fromtimestamp(timestamp, tz=timezone.utc).isoformat(timespec="microseconds")
            size = f"{human_size(entry_stat.st_size):5s}" if human_readable else f"{entry_stat.st_size:10d}"

            print(
                (
                    f"{stat_modestr(entry_stat)} {entry_stat.st_uid:4d} {entry_stat.st_gid:4d} {size} "
                    f"{utc_time} {name}{symlink}"
                ),
                file=stdout,
            )
            return

    hr_spaces = f"{'':5s}" if human_readable else " "
    regular_spaces = f"{'':10s}" if not human_readable else " "

    print(f"??????????    ?    ?{regular_spaces}?{hr_spaces}????-??-??T??:??:??.??????+??:?? {name}", file=stdout)


def ls_scandir(path: fsutil.TargetPath, color: bool = False) -> list[tuple[fsutil.TargetPath, str]]:
    """List a directory for the given path."""
    result = []
    if not path.exists() or not path.is_dir():
        return []

    for file_ in path.iterdir():
        file_type = None
        if color:
            if file_.is_symlink():
                file_type = "ln"
            elif file_.is_dir():
                file_type = "di"
            elif file_.is_file():
                file_type = "fi"

        result.append((file_, fmt_ls_colors(file_type, file_.name) if color else file_.name))

        # If we happen to scan an NTFS filesystem see if any of the
        # entries has an alternative data stream and also list them.
        entry = file_.get()
        if isinstance(entry, LayerFilesystemEntry) and entry.entries.fs.__type__ == "ntfs":
            attrs = entry.lattr()
            for data_stream in attrs.DATA:
                if data_stream.name != "":
                    name = f"{file_.name}:{data_stream.name}"
                    result.append((file_, fmt_ls_colors(file_type, name) if color else name))

    result.sort(key=lambda e: e[0].name)

    return result


def print_ls(
    path: fsutil.TargetPath,
    depth: int,
    stdout: TextIO,
    long_listing: bool = False,
    human_readable: bool = False,
    recursive: bool = False,
    use_ctime: bool = False,
    use_atime: bool = False,
    color: bool = True,
) -> None:
    """Print ls output."""
    subdirs = []

    if path.is_dir():
        contents = ls_scandir(path, color)
    elif path.is_file():
        contents = [(path, path.name)]

    if depth > 0:
        print(f"\n{path!s}:", file=stdout)

    if not long_listing:
        for target_path, name in contents:
            print(name, file=stdout)
            if target_path.is_dir():
                subdirs.append(target_path)
    else:
        if len(contents) > 1:
            print(f"total {len(contents)}", file=stdout)
        for target_path, name in contents:
            try:
                entry = target_path.get()
                entry_stat = entry.lstat()
                show_time = entry_stat.st_mtime
                if use_ctime:
                    show_time = entry_stat.st_ctime
                elif use_atime:
                    show_time = entry_stat.st_atime
            except FileNotFoundError:
                entry = None
                show_time = None
            print_extensive_file_stat_listing(stdout, name, entry, show_time, human_readable)
            if target_path.is_dir():
                subdirs.append(target_path)

    if recursive and subdirs:
        for subdir in subdirs:
            print_ls(subdir, depth + 1, stdout, long_listing, human_readable, recursive, use_ctime, use_atime, color)


def print_stat(path: fsutil.TargetPath, stdout: TextIO, dereference: bool = False) -> None:
    """Print file status."""
    symlink = f"-> {path.readlink()}" if path.is_symlink() else ""
    s = path.stat() if dereference else path.lstat()

    def filetype(path: TargetPath) -> str:
        if path.is_dir():
            return "directory"
        if path.is_symlink():
            return "symbolic link"
        if path.is_file():
            return "regular file"
        return "unknown"

    res = STAT_TEMPLATE.format(
        path=path,
        symlink=symlink,
        size=s.st_size,
        filetype=filetype(path),
        device="?",
        inode=s.st_ino,
        blocks=s.st_blocks if s.st_blocks is not None else "?",
        blksize=s.st_blksize or "?",
        nlink=s.st_nlink,
        modeord=oct(stat.S_IMODE(s.st_mode)),
        modestr=stat_modestr(s),
        uid=s.st_uid,
        gid=s.st_gid,
        atime=datetime.fromtimestamp(s.st_atime, tz=timezone.utc).isoformat(timespec="microseconds"),
        mtime=datetime.fromtimestamp(s.st_mtime, tz=timezone.utc).isoformat(timespec="microseconds"),
        ctime=datetime.fromtimestamp(s.st_ctime, tz=timezone.utc).isoformat(timespec="microseconds"),
        btime=(
            datetime.fromtimestamp(s.st_birthtime, tz=timezone.utc).isoformat(timespec="microseconds")
            if hasattr(s, "st_birthtime") and s.st_birthtime
            else "?"
        ),
    )
    print(res, file=stdout)

    try:
        if (xattr := path.get().attr()) and isinstance(xattr, list) and hasattr(xattr[0], "name"):
            print("  Attr:")
            print_xattr(path.name, xattr, stdout)
    except Exception:
        pass


def print_xattr(basename: str, xattr: list, stdout: TextIO) -> None:
    """Mimics getfattr -d {file} behaviour."""
    if not hasattr(xattr[0], "name"):
        return

    XATTR_TEMPLATE = "# file: {basename}\n{attrs}"
    res = XATTR_TEMPLATE.format(
        basename=basename, attrs="\n".join([f'{attr.name}="{attr.value.decode()}"' for attr in xattr])
    )
    print(res, file=stdout)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.filesystems.ntfs import NtfsFilesystem
from dissect.target.plugins.os.windows.syscache import SyscachePlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_syscache_plugin(target_win: Target, fs_win: VirtualFilesystem) -> None:
    syscache_file = absolute_path("_data/plugins/os/windows/syscache/Syscache.hve")
    fs_win.map_file("System Volume Information/Syscache.hve", syscache_file)

    target_win.add_plugin(SyscachePlugin)

    results = list(target_win.syscache())
    assert len(results) == 401


def test_syscache_plugin_real_mft(target_win: Target, fs_win: VirtualFilesystem) -> None:
    filesystem = NtfsFilesystem(mft=absolute_path("_data/plugins/filesystem/ntfs/mft/mft.raw").open("rb"))

    # We need to change the type of the mocked filesystem, since syscache.py checks for explicit value
    target_win.fs.mounts["sysvol"].__type__ = "ntfs"
    target_win.fs.mounts["sysvol"].ntfs = filesystem.ntfs

    syscache_file = absolute_path("_data/plugins/os/windows/syscache/Syscache-mft.hve")
    fs_win.map_file("System Volume Information/Syscache.hve", syscache_file)

    target_win.add_plugin(SyscachePlugin)

    results = list(target_win.syscache())
    assert len(results) == 401

    filepaths = [entry.path for entry in results]
    assert filepaths.count(None) == 399
    assert "sysvol\\NamelessDirectory\\totally_normal.txt" in filepaths
    assert "sysvol\\text_document.txt" in filepaths



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import PropertyMock, patch

import pytest

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugins.os.unix.linux.proc import ProcPlugin, ProcProcess

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_process(target_linux_users: Target, fs_linux_proc: VirtualFilesystem) -> None:
    target_linux_users.add_plugin(ProcPlugin)

    process = target_linux_users.proc.process(1)
    assert isinstance(process, ProcProcess)
    assert process.pid == 1
    assert process.name == "systemd"
    assert process.parent.name == "swapper"
    assert process.ppid == 0
    assert process.state == "Sleeping"
    assert str(process.runtime) == "1 day, 13:19:27.970000"
    assert process.starttime.isoformat() == "2023-04-03T22:10:54.300000+00:00"

    environ = list(process.environ())
    assert environ[0].variable == "VAR"
    assert environ[0].contents == "1"


def test_process_not_found(target_linux_users: Target, fs_linux_proc: VirtualFilesystem) -> None:
    target_linux_users.add_plugin(ProcPlugin)
    with pytest.raises(ProcessLookupError) as exc:
        target_linux_users.proc.process(404)
    assert str(exc.value) == f"Process with PID 404 could not be found on target: {target_linux_users}"


def test_processes(target_linux_users: Target, fs_linux_proc: VirtualFilesystem) -> None:
    target_linux_users.add_plugin(ProcPlugin)

    for process in target_linux_users.proc.processes():
        assert process.pid in (1, 2, 3, 1337)
        assert process.state in ("Sleeping", "Waking", "Running", "Wakekill")
        assert process.name in ("systemd", "kthread", "acquire", "sshd")
        assert process.starttime.isoformat() == "2023-04-03T22:10:54.300000+00:00"
        assert str(process.runtime) == "1 day, 13:19:27.970000"

        for env in process.environ():
            assert env.variable == "VAR"
            assert env.contents == "1"


def test_processes_without_boottime(target_linux_users: Target, fs_linux_proc: VirtualFilesystem) -> None:
    target_linux_users.add_plugin(ProcPlugin)

    with patch.object(ProcProcess, "_boottime", new_callable=PropertyMock) as mock_boottime:
        mock_boottime.return_value = None

        for process in target_linux_users.proc.processes():
            assert process.pid in (1, 2, 3, 1337)
            assert process.state in ("Sleeping", "Waking", "Running", "Wakekill")
            assert process.name in ("systemd", "kthread", "acquire", "sshd")
            assert process.starttime is None
            assert process.runtime is None

            for env in process.environ():
                assert env.variable == "VAR"
                assert env.contents == "1"


def test_proc_plugin_incompatible(target_linux_users: Target, fs_linux: VirtualFilesystem) -> None:
    with pytest.raises(UnsupportedPluginError, match="No /proc directory found"):
        target_linux_users.add_plugin(ProcPlugin)



from __future__ import annotations

import re
from datetime import datetime
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.remoteaccess.remoteaccess import (
    GENERIC_LOG_RECORD_FIELDS,
    RemoteAccessPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target


RE_LOG = re.compile(
    r"""
        ^
        (?P<date>(\d{2,4}\/)?\d{2}\/\d{2})                  # YYYY/MM/DD or YY/MM/DD or MM/DD
        \s
        (?P<time>\d{2}\:\d{2}\:\d{2}(?:[\.\:]\d{3})?)       # HH:MM:SS or HH:MM:SS.FFF or HH:MM:SS:FFF
        \s+
        (?P<message>.+)
        $
    """,
    re.VERBOSE,
)
RE_START = re.compile(
    r"""
        ^Start\:
        \s+
        (?P<date>\S+)
        \s
        (?P<time>\S+)
        (
            \s
            \((?P<timezone>\S+)\)                           # UTC+2:00
        )?
        $
    """,
    re.VERBOSE,
)


class TeamViewerPlugin(RemoteAccessPlugin):
    """TeamViewer client plugin.

    References:
        - https://teamviewer.com/en/global/support/knowledge-base/teamviewer-classic/contact-support/find-your-log-files
        - https://www.systoolsgroup.com/forensics/teamviewer/
        - https://benleeyr.wordpress.com/2020/05/19/teamviewer-forensics-tested-on-v15/
    """

    __namespace__ = "teamviewer"

    SYSTEM_GLOBS = (
        "sysvol/Program Files/TeamViewer/*.log",
        "sysvol/Program Files (x86)/TeamViewer/*.log",
        "/var/log/teamviewer*/*.log",
    )

    USER_GLOBS = (
        "AppData/Roaming/TeamViewer/teamviewer*_logfile.log",
        "Library/Logs/TeamViewer/teamviewer*_logfile*.log",
    )

    RemoteAccessLogRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "remoteaccess/teamviewer/log", GENERIC_LOG_RECORD_FIELDS
    )

    def __init__(self, target: Target):
        super().__init__(target)

        self.logfiles: set[tuple[str, UserDetails | None]] = set()

        # Find system service log files.
        for log_glob in self.SYSTEM_GLOBS:
            for logfile in self.target.fs.glob(log_glob):
                self.logfiles.add((logfile, None))

        # Find user log files.
        for user_details in self.target.user_details.all_with_home():
            for log_glob in self.USER_GLOBS:
                for logfile in user_details.home_path.glob(log_glob):
                    self.logfiles.add((logfile, user_details))

    def check_compatible(self) -> None:
        if not len(self.logfiles):
            raise UnsupportedPluginError("No Teamviewer logs found on target")

    @export(record=RemoteAccessLogRecord)
    def logs(self) -> Iterator[RemoteAccessLogRecord]:
        """Yield TeamViewer client logs.

        TeamViewer is a commercial remote desktop application. An adversary may use it to gain persistence on a system.
        """
        target_tz = self.target.datetime.tzinfo

        for logfile, user_details in self.logfiles:
            logfile = self.target.fs.path(logfile)

            start_date = None
            for line in logfile.open("rt", errors="replace"):
                if not (line := line.strip()) or line.startswith("# "):
                    continue

                if line.startswith("Start:"):
                    try:
                        start_date = parse_start(line)
                    except Exception as e:
                        self.target.log.warning("Failed to parse Start message %r in %s", line, logfile)
                        self.target.log.debug("", exc_info=e)

                    continue

                if not (match := RE_LOG.search(line)):
                    self.target.log.warning("Skipping TeamViewer log line %r in %s", line, logfile)
                    continue

                log = match.groupdict()
                date = log["date"]
                time = log["time"]

                # Older TeamViewer versions first mention the start time and then leave out the year,
                # so we have to correct for the missing year in date.
                if date.count("/") == 1:
                    if not start_date:
                        self.target.log.warning("Missing year in log line, skipping line %r in %s", line, logfile)
                        continue
                    date = f"{start_date.year}/{log['date']}"

                # Correct for year if short notation for 2000 is used
                if date.count("/") == 2 and len(date.split("/")[0]) == 2:
                    date = "20" + date

                # Correct for ``:`` separator of milliseconds
                if time.count(":") == 3:
                    hms, _, ms = time.rpartition(":")
                    time = f"{hms}.{ms}"

                # Convert milliseconds to microseconds
                if "." in time:
                    hms, _, ms = time.rpartition(".")
                    time = f"{hms}.{ms:0<6}"
                else:
                    time += ".000000"

                try:
                    timestamp = datetime.strptime(f"{date} {time}", "%Y/%m/%d %H:%M:%S.%f").replace(
                        tzinfo=start_date.tzinfo if start_date else target_tz
                    )
                except Exception as e:
                    self.target.log.warning("Unable to parse timestamp %r in file %s", line, logfile)
                    self.target.log.debug("", exc_info=e)
                    timestamp = 0

                yield self.RemoteAccessLogRecord(
                    ts=timestamp,
                    message=log.get("message"),
                    source=logfile,
                    _target=self.target,
                    _user=user_details.user if user_details else None,
                )


def parse_start(line: str) -> datetime | None:
    """TeamViewer ``Start`` messages can be formatted in different ways
    and might contain the timezone offset of all timestamps.

    .. code-block::

        Start: 2021/11/11 12:34:56
        Start: 2024/12/31 01:02:03.123 (UTC+2:00)
    """

    if match := RE_START.search(line):
        dt = match.groupdict()

        # Drop milliseconds
        if "." in dt["time"]:
            dt["time"] = dt["time"].rsplit(".")[0]

        # Format timezone, e.g. "UTC+2:00" to "UTC+0200"
        if dt["timezone"]:
            name, operator, amount = re.split(r"(\+|\-)", dt["timezone"])
            amount = int(amount.replace(":", ""))
            dt["timezone"] = f"{name}{operator}{amount:0>4d}"

        return datetime.strptime(  # noqa: DTZ007
            f"{dt['date']} {dt['time']}" + (f" {dt['timezone']}" if dt["timezone"] else ""),
            "%Y/%m/%d %H:%M:%S" + (" %Z%z" if dt["timezone"] else ""),
        )
    return None
